#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Any, Sequence

# Bokeh imports
from ..core.has_props import abstract
from ..model import Model

@abstract
@dataclass(init=False)
class Filter(Model):

    def __invert__(self) -> Filter: ...

    def __and__(self, other: Filter) -> Filter: ...

    def __or__(self, other: Filter) -> Filter: ...

    def __sub__(self, other: Filter) -> Filter: ...

    def __xor__(self, other: Filter) -> Filter: ...

@dataclass
class AllIndices(Filter):
    ...

@dataclass
class InversionFilter(Filter):

    operand: Filter = ...

@abstract
@dataclass(init=False)
class CompositeFilter(Filter):

    operands: Sequence[Filter] = ...

@dataclass
class IntersectionFilter(CompositeFilter):
    ...

@dataclass
class UnionFilter(CompositeFilter):
    ...

@dataclass
class DifferenceFilter(CompositeFilter):
    ...

@dataclass
class SymmetricDifferenceFilter(CompositeFilter):
    ...

@dataclass
class IndexFilter(Filter):

    indices: Sequence[int] | None = ...

@dataclass
class BooleanFilter(Filter):

    booleans: Sequence[bool] | None = ...

@dataclass
class GroupFilter(Filter):

    column_name: str = ...

    group: Any = ...

@dataclass
class CustomJSFilter(Filter):

    args: dict[str, Any] = ...

    code: str = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Provide a collection of general utilities useful for implementing Bokeh
functionality.

.. _bokeh.util.browser:

``bokeh.util.browser``
----------------------

.. automodule:: bokeh.util.browser
  :members:

.. _bokeh.util.callback_manager:

``bokeh.util.callback_manager``
-------------------------------

.. automodule:: bokeh.util.callback_manager
  :members:

.. _bokeh.util.compiler:

``bokeh.util.compiler``
-----------------------

.. automodule:: bokeh.util.compiler
  :members:

.. _bokeh.util.dependencies:

``bokeh.util.dependencies``
---------------------------

.. automodule:: bokeh.util.dependencies
  :members:

.. _bokeh.util.deprecation:

``bokeh.util.deprecation``
--------------------------

.. automodule:: bokeh.util.deprecation

.. _bokeh.util.functions:

``bokeh.util.functions``
------------------------

.. automodule:: bokeh.util.functions
  :members:

``bokeh.util.hex``
------------------

.. automodule:: bokeh.util.hex
  :members:

.. _bokeh.util.info:

``bokeh.util.info``
----------------------

.. automodule:: bokeh.util.info
  :members:

.. _bokeh.util.logconfig:

``bokeh.util.logconfig``
------------------------

.. automodule:: bokeh.util.logconfig
  :members:

.. _bokeh.util.options:

``bokeh.util.options``
----------------------

.. automodule:: bokeh.util.options
  :members:

.. _bokeh.util.paths:

``bokeh.util.paths``
--------------------

.. automodule:: bokeh.util.paths
  :members:

.. _bokeh.util.serialization:

``bokeh.util.serialization``
----------------------------

.. automodule:: bokeh.util.serialization
  :members:

.. _bokeh.util.token:

``bokeh.util.token``
--------------------

.. automodule:: bokeh.util.token
  :members:

.. _bokeh.util.strings:

``bokeh.util.strings``
----------------------

.. automodule:: bokeh.util.strings
  :members:

.. _bokeh.util.tornado:

``bokeh.util.tornado``
----------------------

.. automodule:: bokeh.util.tornado
  :members:

  .. _bokeh.util.terminal:

``bokeh.util.terminal``
-----------------------

.. automodule:: bokeh.util.terminal
  :members:

.. _bokeh.util.version:

``bokeh.util.version``
----------------------

.. automodule:: bokeh.util.version
  :members:

.. _bokeh.util.warnings:

``bokeh.util.warnings``
-----------------------

.. automodule:: bokeh.util.warnings
  :members:

'''



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Sequence

# Bokeh imports
from ..core.has_props import abstract
from ..model import Model
from .expressions import CoordinateTransform

@abstract
@dataclass(init=False)
class LayoutProvider(Model):

    @property
    def node_coordinates(self) -> NodeCoordinates: ...

    @property
    def edge_coordinates(self) -> EdgeCoordinates: ...

@dataclass
class StaticLayoutProvider(LayoutProvider):

    graph_layout: dict[int | str, Sequence[float]] = ...

@abstract
@dataclass(init=False)
class GraphCoordinates(CoordinateTransform):

    layout: LayoutProvider = ...

@dataclass
class NodeCoordinates(GraphCoordinates):
    ...

@dataclass
class EdgeCoordinates(GraphCoordinates):
    ...

@abstract
@dataclass(init=False)
class GraphHitTestPolicy(Model):
    ...

@dataclass
class EdgesOnly(GraphHitTestPolicy):
    ...

@dataclass
class NodesOnly(GraphHitTestPolicy):
    ...

@dataclass
class NodesAndLinkedEdges(GraphHitTestPolicy):
    ...

@dataclass
class EdgesAndLinkedNodes(GraphHitTestPolicy):
    ...

@dataclass
class NodesAndAdjacentNodes(GraphHitTestPolicy):
    ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Any, Sequence

# Bokeh imports
from .._specs import AngleSpec, NumberSpec
from ..core.enums import DirectionType as Direction
from ..core.has_props import abstract
from ..model import Model

@abstract
@dataclass(init=False)
class Expression(Model):
    ...

@dataclass
class CustomJSExpr(Expression):

    args: dict[str, Any] = ...

    code: str = ...

@dataclass
class CumSum(Expression):

    field: str = ...

    include_zero: bool = ...

@dataclass
class Stack(Expression):

    fields: Sequence[str] = ...

@abstract
@dataclass(init=False)
class ScalarExpression(Model):
    ...

@dataclass
class Minimum(ScalarExpression):

    field: str
    initial: float | None

@dataclass
class Maximum(ScalarExpression):

    field: str
    initial: float | None

@abstract
@dataclass(init=False)
class CoordinateTransform(Expression):

    @property
    def x(self) -> XComponent: ...

    @property
    def y(self) -> YComponent: ...

@dataclass
class PolarTransform(CoordinateTransform):

    radius: NumberSpec = ...

    angle: AngleSpec = ...

    direction: Direction = ...

@abstract
@dataclass(init=False)
class XYComponent(Expression):

    transform: CoordinateTransform = ...

@dataclass
class XComponent(XYComponent):
    ...

@dataclass
class YComponent(XYComponent):
    ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Callable

# Bokeh imports
from ...core.enums import ButtonTypeType as ButtonType
from ...core.has_props import HasProps, abstract
from ...util.callback_manager import EventCallback
from ..callbacks import Callback
from ..dom import DOMNode
from ..ui.icons import Icon
from ..ui.tooltips import Tooltip
from .widget import Widget

@abstract
@dataclass(init=False)
class ButtonLike(HasProps):

    button_type: ButtonType = ...

@abstract
@dataclass(init=False)
class AbstractButton(Widget, ButtonLike):

    label: DOMNode | str = ...

    icon: Icon | None = ...

@dataclass
class Button(AbstractButton):

    def on_click(self, handler: EventCallback) -> None: ...

    def js_on_click(self, handler: Callback) -> None: ...

@dataclass
class Toggle(AbstractButton):

    active: bool = ...

    def on_click(self, handler: Callable[[bool], None]) -> None: ...

    def js_on_click(self, handler: Callback) -> None: ...

@dataclass
class Dropdown(AbstractButton):

    split: bool = ...

    menu: list[str | tuple[str, str | Callback] | None] = ...

    def on_click(self, handler: EventCallback) -> None: ...

    def js_on_click(self, handler: Callback) -> None: ...

@dataclass
class HelpButton(AbstractButton):

    tooltip: Tooltip = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Literal

# Bokeh imports
from ..._types import (
    Date,
    Datetime,
    Positive,
    Time,
)
from ...core.enums import CalendarPositionType as CalendarPosition
from ...core.has_props import HasProps, abstract
from .inputs import InputWidget

@abstract
@dataclass(init=False)
class PickerBase(InputWidget):

    position: CalendarPosition = ...

    inline: bool = ...

@abstract
@dataclass(init=False)
class TimeCommon(HasProps):

    hour_increment: Positive[int] = ...

    minute_increment: Positive[int] = ...

    second_increment: Positive[int] = ...

    seconds: bool = ...

    clock: Literal["12h", "24h"] = ...

@dataclass
class TimePicker(PickerBase, TimeCommon):

    value: Time | None = ...

    time_format: str = ...

    min_time: Time | None = ...

    max_time: Time | None = ...

@abstract
@dataclass(init=False)
class DateCommon(HasProps):

    disabled_dates: list[Date | tuple[Date, Date]] | None = ...

    enabled_dates: list[Date | tuple[Date, Date]] | None = ...

    date_format: str = ...

@abstract
@dataclass(init=False)
class BaseDatePicker(PickerBase, DateCommon):

    min_date: Date | None = ...

    max_date: Date | None = ...

@dataclass
class DatePicker(BaseDatePicker):

    value: Date | None = ...

@dataclass
class DateRangePicker(BaseDatePicker):

    value: tuple[Date, Date] | None = ...

@dataclass
class MultipleDatePicker(BaseDatePicker):

    value: list[Date] = ...

    separator: str = ...

@abstract
@dataclass(init=False)
class BaseDatetimePicker(PickerBase, DateCommon, TimeCommon):

    min_date: Datetime | Date | None = ...

    max_date: Datetime | Date | None = ...

@dataclass
class DatetimePicker(BaseDatetimePicker):

    value: Datetime | None = ...

@dataclass
class DatetimeRangePicker(BaseDatetimePicker):

    value: tuple[Datetime, Datetime] | None = ...

@dataclass
class MultipleDatetimePicker(BaseDatetimePicker):

    value: list[Datetime] = ...

    separator: str = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..tiles import TileSource
from .renderer import Renderer

@dataclass
class TileRenderer(Renderer):

    tile_source: TileSource = ...

    alpha: float = ...

    smoothing: bool = ...

    render_parents: bool = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
'''

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import Any

# Bokeh imports
from ....core.enums import (
    AngleUnits,
    CoordinateUnits,
    FontStyle,
    TextAlign,
    VerticalAlign,
)
from ....core.properties import (
    Alpha,
    Angle,
    AngleSpec,
    Color,
    CoordinateLike,
    Enum,
    Float,
    Include,
    Nullable,
    NullStringSpec,
    NumberSpec,
    Override,
    Required,
    String,
    field,
)
from ....core.property_aliases import BorderRadius, Padding
from ....core.property_mixins import (
    FillProps,
    LineProps,
    ScalarFillProps,
    ScalarHatchProps,
    ScalarLineProps,
    ScalarTextProps,
    TextProps,
)
from ..annotation import DataAnnotation
from .html_annotation import HTMLAnnotation

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    "HTMLLabel",
    "HTMLLabelSet",
    "HTMLTitle",
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

class HTMLTextAnnotation(HTMLAnnotation):

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    padding = Padding(default=0, help="""
    Extra space between the text of a label and its bounding box (border).

    .. note::
        This property is experimental and may change at any point.
    """)

    border_radius = BorderRadius(default=0, help="""
    Allows label's box to have rounded corners. For the best results, it
    should be used in combination with ``padding``.

    .. note::
        This property is experimental and may change at any point.
    """)

    background_props = Include(ScalarFillProps, prefix="background", help="""
    The {prop} values for the text bounding box.
    """)

    background_hatch_props = Include(ScalarHatchProps, prefix="background", help="""
    The {prop} values for the text bounding box.
    """)

    border_props = Include(ScalarLineProps, prefix="border", help="""
    The {prop} values for the text bounding box.
    """)

    background_fill_color = Override(default=None)

    background_hatch_color = Override(default=None)

    border_line_color = Override(default=None)

class HTMLLabel(HTMLTextAnnotation):
    ''' Render a single HTML label as an annotation.

    ``Label`` will render a single text label at given ``x`` and ``y``
    coordinates, which can be in either screen (pixel) space, or data (axis
    range) space.

    The label can also be configured with a screen space offset from ``x`` and
    ``y``, by using the ``x_offset`` and ``y_offset`` properties.

    Additionally, the label can be rotated with the ``angle`` property.

    There are also standard text, fill, and line properties to control the
    appearance of the text, its background, as well as the rectangular bounding
    box border.

    See :ref:`ug_basic_annotations_labels` for information on plotting labels.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    x = Required(CoordinateLike, help="""
    The x-coordinate in screen coordinates to locate the text anchors.
    """)

    x_units = Enum(CoordinateUnits, default='data', help="""
    The unit type for the x attribute. Interpreted as |data units| by
    default.
    """)

    y = Required(CoordinateLike, help="""
    The y-coordinate in screen coordinates to locate the text anchors.
    """)

    y_units = Enum(CoordinateUnits, default='data', help="""
    The unit type for the y attribute. Interpreted as |data units| by
    default.
    """)

    text = String(default="", help="""
    The text value to render.
    """)

    angle = Angle(default=0, help="""
    The angle to rotate the text, as measured from the horizontal.
    """)

    angle_units = Enum(AngleUnits, default='rad', help="""
    Acceptable values for units are ``"rad"`` and ``"deg"``
    """)

    x_offset = Float(default=0, help="""
    Offset value to apply to the x-coordinate.

    This is useful, for instance, if it is desired to "float" text a fixed
    distance in |screen units| from a given data position.
    """)

    y_offset = Float(default=0, help="""
    Offset value to apply to the y-coordinate.

    This is useful, for instance, if it is desired to "float" text a fixed
    distance in |screen units| from a given data position.
    """)

    text_props = Include(ScalarTextProps, help="""
    The {prop} values for the text.
    """)

class HTMLLabelSet(HTMLAnnotation, DataAnnotation):
    ''' Render multiple text labels as annotations.

    ``LabelSet`` will render multiple text labels at given ``x`` and ``y``
    coordinates, which can be in either screen (pixel) space, or data (axis
    range) space. In this case (as opposed to the single ``Label`` model),
    ``x`` and ``y`` can also be the name of a column from a
    :class:`~bokeh.models.sources.ColumnDataSource`, in which case the labels
    will be "vectorized" using coordinate values from the specified columns.

    The label can also be configured with a screen space offset from ``x`` and
    ``y``, by using the ``x_offset`` and ``y_offset`` properties. These offsets
    may be vectorized by giving the name of a data source column.

    Additionally, the label can be rotated with the ``angle`` property (which
    may also be a column name.)

    There are also standard text, fill, and line properties to control the
    appearance of the text, its background, as well as the rectangular bounding
    box border.

    The data source is provided by setting the ``source`` property.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    x = NumberSpec(default=field("x"), help="""
    The x-coordinates to locate the text anchors.
    """)

    x_units = Enum(CoordinateUnits, default='data', help="""
    The unit type for the ``xs`` attribute. Interpreted as |data units| by
    default.
    """)

    y = NumberSpec(default=field("y"), help="""
    The y-coordinates to locate the text anchors.
    """)

    y_units = Enum(CoordinateUnits, default='data', help="""
    The unit type for the ``ys`` attribute. Interpreted as |data units| by
    default.
    """)

    text = NullStringSpec(default=field("text"), help="""
    The text values to render.
    """)

    angle = AngleSpec(default=0, help="""
    The angles to rotate the text, as measured from the horizontal.
    """)

    x_offset = NumberSpec(default=0, help="""
    Offset values to apply to the x-coordinates.

    This is useful, for instance, if it is desired to "float" text a fixed
    distance in |screen units| from a given data position.
    """)

    y_offset = NumberSpec(default=0, help="""
    Offset values to apply to the y-coordinates.

    This is useful, for instance, if it is desired to "float" text a fixed
    distance in |screen units| from a given data position.
    """)

    text_props = Include(TextProps, help="""
    The {prop} values for the text.
    """)

    background_props = Include(FillProps, prefix="background", help="""
    The {prop} values for the text bounding box.
    """)

    background_fill_color = Override(default=None)

    border_props = Include(LineProps, prefix="border", help="""
    The {prop} values for the text bounding box.
    """)

    border_line_color = Override(default=None)

class HTMLTitle(HTMLTextAnnotation):
    ''' Render a single title box as an annotation.

    See :ref:`ug_basic_annotations_titles` for information on plotting titles.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    text = String(default="", help="""
    The text value to render.
    """)

    vertical_align = Enum(VerticalAlign, default='bottom', help="""
    Alignment of the text in its enclosing space, *across* the direction of the text.
    """)

    align = Enum(TextAlign, default='left', help="""
    Alignment of the text in its enclosing space, *along* the direction of the text.
    """)

    text_line_height = Float(default=1.0, help="""
    How much additional space should be allocated for the title. The value is provided
    as a number, but should be treated as a percentage of font size. The default is
    100%, which means no additional space will be used.
    """)

    offset = Float(default=0, help="""
    Offset the text by a number of pixels (can be positive or negative). Shifts the text in
    different directions based on the location of the title:

        * above: shifts title right
        * right: shifts title down
        * below: shifts title right
        * left: shifts title up

    """)

    standoff = Float(default=10, help="""
    """)

    text_font = String(default="helvetica", help="""
    Name of a font to use for rendering text, e.g., ``'times'``,
    ``'helvetica'``.

    """)

    text_font_size = String(default="13px")

    text_font_style = Enum(FontStyle, default="bold", help="""
    A style to use for rendering text.

    Acceptable values are:

    - ``'normal'`` normal text
    - ``'italic'`` *italic text*
    - ``'bold'`` **bold text**

    """)

    text_color = Color(default="#444444", help="""
    A color to use to fill text with.
    """)

    text_outline_color = Nullable(Color, default=None, help="""
    A color to use to fill text with.
    """)

    text_alpha = Alpha(help="""
    An alpha value to use to fill text with.
    """)

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Sequence

# Bokeh imports
from ..._specs import CoordinateSpec
from ..._types import (
    Coordinate,
    CoordinateLike,
    NonNegative,
    Positive,
)
from ...core.enums import (
    CoordinateUnitsType as CoordinateUnits,
    DimensionType as Dimension,
    MovableType as Movable,
    ResizableType as Resizable,
)
from ...core.property_aliases import BorderRadius
from ...core.property_mixins import (
    LineProps,
    ScalarAboveFillProps,
    ScalarAboveHatchProps,
    ScalarBelowFillProps,
    ScalarBelowHatchProps,
    ScalarFillProps,
    ScalarHatchProps,
    ScalarHoverFillProps,
    ScalarHoverHatchProps,
    ScalarHoverLineProps,
    ScalarLineProps,
)
from ...model import Model
from ..nodes import BoxNodes
from .annotation import Annotation, DataAnnotation
from .arrows import ArrowHead

@dataclass
class AreaVisuals(Model, ScalarLineProps, ScalarFillProps, ScalarHatchProps,
                  ScalarHoverLineProps, ScalarHoverFillProps, ScalarHoverHatchProps):
    ...

@dataclass
class BoxInteractionHandles(Model):

    all: AreaVisuals = ...

    move: AreaVisuals | None = ...
    resize: AreaVisuals | None = ...

    sides: AreaVisuals | None = ...
    corners: AreaVisuals | None = ...

    left: AreaVisuals | None = ...
    right: AreaVisuals | None = ...
    top: AreaVisuals | None = ...
    bottom: AreaVisuals | None = ...

    top_left: AreaVisuals | None = ...
    top_right: AreaVisuals | None = ...
    bottom_left: AreaVisuals | None = ...
    bottom_right: AreaVisuals | None = ...

@dataclass
class BoxAnnotation(Annotation, AreaVisuals):

    left: Coordinate = ...

    right: Coordinate = ...

    top: Coordinate = ...

    bottom: Coordinate = ...

    left_units: CoordinateUnits = ...

    right_units: CoordinateUnits = ...

    top_units: CoordinateUnits = ...

    bottom_units: CoordinateUnits = ...

    left_limit: Coordinate | None = ...

    right_limit: Coordinate | None = ...

    top_limit: Coordinate | None = ...

    bottom_limit: Coordinate | None = ...

    min_width: NonNegative[float] = ...

    min_height: NonNegative[float] = ...

    max_width: Positive[float] = ...

    max_height: Positive[float] = ...

    border_radius: BorderRadius = ...

    editable: bool = ...

    resizable: Resizable = ...

    movable: Movable = ...

    symmetric: bool = ...

    use_handles: bool = ...

    handles: BoxInteractionHandles | AreaVisuals = ...

    inverted: bool = ...

    @property
    def nodes(self) -> BoxNodes: ...

@dataclass
class Band(DataAnnotation, ScalarLineProps, ScalarFillProps, ScalarHatchProps):

    lower: CoordinateSpec = ...

    upper: CoordinateSpec = ...

    base: CoordinateSpec = ...

    dimension: Dimension = ...

@dataclass
class PolyAnnotation(Annotation, ScalarLineProps, ScalarFillProps, ScalarHatchProps,
                     ScalarHoverLineProps, ScalarHoverFillProps, ScalarHoverHatchProps):

    xs: Sequence[CoordinateLike] = ...

    xs_units: CoordinateUnits = ...

    ys: Sequence[CoordinateLike] = ...

    ys_units: CoordinateUnits = ...

    editable: bool = ...

@dataclass
class Slope(Annotation, ScalarLineProps, ScalarAboveFillProps, ScalarAboveHatchProps, ScalarBelowFillProps, ScalarBelowHatchProps):

    gradient: float | None = ...

    y_intercept: float | None = ...

@dataclass
class Span(Annotation, ScalarLineProps, ScalarHoverLineProps):

    location: CoordinateLike | None = ...

    location_units: CoordinateUnits = ...

    dimension: Dimension = ...

    editable: bool = ...

@dataclass
class Whisker(DataAnnotation, LineProps ):

    lower: CoordinateSpec = ...

    lower_head: ArrowHead | None = ...

    upper: CoordinateSpec = ...

    upper_head: ArrowHead | None = ...

    base: CoordinateSpec = ...

    dimension: Dimension = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from typing import (
    Generic,
    Never,
    NotRequired,
    Sequence,
    TypeAlias,
    TypedDict,
    TypeVar,
)

# External imports
import numpy as np
import numpy.typing as npt

# Bokeh imports
from ._types import (
    Color,
    Datetime,
    FontSize,
    FontStyle,
    NonNegative,
    TimeDelta,
)
from .core.enums import (
    AngleUnitsType as AngleUnits,
    CoordinateUnitsType as CoordinateUnits,
    HatchPatternType as HatchPattern,
    LineCapType as LineCap,
    LineJoinType as LineJoin,
    MarkerTypeType as MarkerType,
    OutlineShapeNameType as OutlineShapeName,
    SpatialUnitsType as SpatialUnits,
    TextAlignType as TextAlign,
    TextBaselineType as TextBaseline,
)
from .core.property.vectorization import Expr, Field, Value
from .core.property.visual import DashPatternType as DashPattern
from .core.property_aliases import TextAnchorType as TextAnchor
from .models.expressions import Expression
from .models.transforms import Transform

FieldName: TypeAlias = str

ValueType = TypeVar("ValueType")
UnitsType = TypeVar("UnitsType")

class ValueDict(TypedDict, Generic[ValueType, UnitsType]):
    value: ValueType
    transform: NotRequired[Transform]
    units: NotRequired[UnitsType]

class FieldDict(TypedDict, Generic[ValueType, UnitsType]):
    field: FieldName
    transform: NotRequired[Transform]
    units: NotRequired[UnitsType]

class ExprDict(TypedDict, Generic[ValueType, UnitsType]):
    expr: Expression
    transform: NotRequired[Transform]
    units: NotRequired[UnitsType]

VectorInit: TypeAlias = Value[ValueType] | Field | Expr
VectorDict: TypeAlias = ValueDict[ValueType, UnitsType] | FieldDict[ValueType, UnitsType] | ExprDict[ValueType, UnitsType]
VectorLike: TypeAlias = VectorInit[ValueType] | VectorDict[ValueType, UnitsType]

Vectorized: TypeAlias = FieldName | ValueType | VectorLike[ValueType, UnitsType]

IntArray: TypeAlias = npt.NDArray[np.integer]
FloatArray: TypeAlias = npt.NDArray[np.floating]
NumberArray: TypeAlias = FloatArray | npt.NDArray[np.datetime64] | npt.NDArray[np.timedelta64]
Number1dArray: TypeAlias = NumberArray # TODO shape
Number2dArray: TypeAlias = NumberArray # TODO shape
Number3dArray: TypeAlias = NumberArray # TODO shape
StringArray: TypeAlias = npt.NDArray[np.str_]

T = TypeVar("T")
U = TypeVar("U")

DataSpec: TypeAlias = Vectorized[T, Never]
UnitsSpec: TypeAlias = Vectorized[T, U]

IntVal: TypeAlias = int
IntSpec: TypeAlias = DataSpec[IntVal]
IntArg: TypeAlias = FieldName | IntVal | IntSpec | Sequence[IntVal] | IntArray

FloatVal: TypeAlias = float
FloatSpec: TypeAlias = DataSpec[FloatVal]
FloatArg: TypeAlias = FieldName | FloatVal | FloatSpec | Sequence[FloatVal] | FloatArray

NumberVal: TypeAlias = float | Datetime | TimeDelta
NumberSpec: TypeAlias = DataSpec[NumberVal]
NumberArg: TypeAlias = FieldName | NumberVal | NumberSpec | Sequence[NumberVal] | NumberArray

Number1dVal: TypeAlias = Sequence[float | Datetime | TimeDelta]
Number1dSpec: TypeAlias = DataSpec[Number1dVal]
Number1dArg: TypeAlias = FieldName | Number1dVal | Number1dSpec | Sequence[Number1dVal] | Number1dArray

Number2dVal: TypeAlias = Sequence[Sequence[float | Datetime | TimeDelta]]
Number2dSpec: TypeAlias = DataSpec[Number2dVal]
Number2dArg: TypeAlias = FieldName | Number2dVal | Number2dSpec | Sequence[Number2dVal] | Number2dArray

Number3dVal: TypeAlias = Sequence[Sequence[Sequence[float | Datetime | TimeDelta]]]
Number3dSpec: TypeAlias = DataSpec[Number3dVal]
Number3dArg: TypeAlias = FieldName | Number3dVal | Number3dSpec | Sequence[Number3dVal] | Number3dArray

Image2dVal: TypeAlias = Number2dArray
Image2dSpec: TypeAlias = DataSpec[Image2dVal]
Image2dArg: TypeAlias = FieldName | Image2dSpec | Sequence[Image2dVal]

Image3dVal: TypeAlias = Number3dArray
Image3dSpec: TypeAlias = DataSpec[Image3dVal]
Image3dArg: TypeAlias = FieldName | Image3dSpec | Sequence[Image3dVal]

SizeVal: TypeAlias = NonNegative[float] | Datetime | TimeDelta
SizeSpec: TypeAlias = DataSpec[SizeVal]
SizeArg: TypeAlias = FieldName | SizeVal | Sequence[SizeVal] | SizeSpec | NumberArray

AlphaVal: TypeAlias = FloatVal
AlphaSpec: TypeAlias = FloatSpec
AlphaArg: TypeAlias = FloatArg

ColorVal: TypeAlias = Color | None
ColorSpec: TypeAlias = DataSpec[Color | None]
ColorArg: TypeAlias = FieldName | ColorVal | Sequence[ColorVal] | ColorSpec | npt.NDArray[np.uint8] # TODO

StringVal: TypeAlias = str
StringSpec: TypeAlias = DataSpec[StringVal]
StringArg: TypeAlias = FieldName | StringVal | Sequence[StringVal] | StringSpec | StringArray

NullStringVal: TypeAlias = StringVal | None
NullStringSpec: TypeAlias = DataSpec[NullStringVal]
NullStringArg: TypeAlias = FieldName | NullStringVal | Sequence[NullStringVal] | NullStringSpec | StringArray

FontSizeVal: TypeAlias = FontSize
FontSizeSpec: TypeAlias = DataSpec[FontSizeVal]
FontSizeArg: TypeAlias = FieldName | FontSizeVal | Sequence[FontSizeVal] | FontSizeSpec | StringArray

FontStyleVal: TypeAlias = FontStyle
FontStyleSpec: TypeAlias = DataSpec[FontStyleVal]
FontStyleArg: TypeAlias = FieldName | FontStyleVal | Sequence[FontStyleVal] | FontStyleSpec | StringArray

TextAlignVal: TypeAlias = TextAlign
TextAlignSpec: TypeAlias = DataSpec[TextAlignVal]
TextAlignArg: TypeAlias = FieldName | TextAlignVal | Sequence[TextAlignVal] | TextAlignSpec | StringArray

TextBaselineVal: TypeAlias = TextBaseline
TextBaselineSpec: TypeAlias = DataSpec[TextBaselineVal]
TextBaselineArg: TypeAlias = FieldName | TextBaselineVal | Sequence[TextBaselineVal] | TextBaselineSpec | StringArray

LineJoinVal: TypeAlias = LineJoin
LineJoinSpec: TypeAlias = DataSpec[LineJoinVal]
LineJoinArg: TypeAlias = FieldName | LineJoinVal | Sequence[LineJoinVal] | LineJoinSpec | StringArray

LineCapVal: TypeAlias = LineCap
LineCapSpec: TypeAlias = DataSpec[LineCapVal]
LineCapArg: TypeAlias = FieldName | LineCapVal | Sequence[LineCapVal] | LineCapSpec | StringArray

DashPatternVal: TypeAlias = DashPattern
DashPatternSpec: TypeAlias = DataSpec[DashPatternVal]
DashPatternArg: TypeAlias = FieldName | DashPatternVal | Sequence[DashPatternVal] | DashPatternSpec | StringArray

HatchPatternVal: TypeAlias = HatchPattern | None
HatchPatternSpec: TypeAlias = DataSpec[HatchPatternVal]
HatchPatternArg: TypeAlias = FieldName | HatchPatternVal | Sequence[HatchPatternVal] | HatchPatternSpec | StringArray

MarkerVal: TypeAlias = MarkerType | str
MarkerSpec: TypeAlias = DataSpec[MarkerVal]
MarkerArg: TypeAlias = FieldName | MarkerVal | Sequence[MarkerVal] | MarkerSpec | StringArray

TextAnchorVal: TypeAlias = TextAnchor
TextAnchorSpec: TypeAlias = DataSpec[TextAnchorVal]
TextAnchorArg: TypeAlias = FieldName | TextAnchorVal | Sequence[TextAnchorVal] | TextAnchorSpec | StringArray

OutlineShapeNameVal: TypeAlias = OutlineShapeName
OutlineShapeNameSpec: TypeAlias = DataSpec[OutlineShapeNameVal]
OutlineShapeNameArg: TypeAlias = FieldName | OutlineShapeNameVal | Sequence[OutlineShapeNameVal] | OutlineShapeNameSpec | StringArray

AngleVal: TypeAlias = float
AngleSpec: TypeAlias = UnitsSpec[AngleVal, AngleUnits]
AngleArg: TypeAlias = FieldName | AngleVal | Sequence[AngleVal] | AngleSpec | FloatArray

CoordinateVal: TypeAlias = float | Datetime | TimeDelta
CoordinateSpec: TypeAlias = UnitsSpec[float | Datetime | TimeDelta, CoordinateUnits]
CoordinateArg: TypeAlias = FieldName | CoordinateVal | CoordinateSpec | Sequence[CoordinateVal] | NumberArray

DistanceVal: TypeAlias = NonNegative[float] | Datetime | TimeDelta
DistanceSpec: TypeAlias = UnitsSpec[DistanceVal, SpatialUnits]
DistanceArg: TypeAlias = FieldName | DistanceVal | DistanceSpec | Sequence[DistanceVal] | NumberArray

NullDistanceVal: TypeAlias = DistanceVal | None
NullDistanceSpec: TypeAlias = UnitsSpec[NullDistanceVal, SpatialUnits]
NullDistanceArg: TypeAlias = FieldName | NullDistanceVal | NullDistanceSpec | Sequence[NullDistanceVal] | NumberArray



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Provide a Bokeh Application Handler to build up documents by compiling
and executing Python source code.

This Handler is used by the Bokeh server command line tool to build
applications that run off scripts and notebooks.

.. code-block:: python

    def make_doc(doc: Document):
        # do work to modify the document, add plots, widgets, etc.
        return doc

    app = Application(FunctionHandler(make_doc))

    server = Server({'/bkapp': app}, io_loop=IOLoop.current())
    server.start()

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from contextlib import contextmanager
from os.path import basename, splitext
from types import ModuleType
from typing import Any, Callable, ClassVar

# Bokeh imports
from ...core.types import PathLike
from ...document import Document
from ...io.doc import curdoc, patch_curdoc
from .code_runner import CodeRunner
from .handler import Handler

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'CodeHandler',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

class CodeHandler(Handler):
    ''' Run source code which modifies a Document

    '''

    # These functions, if present in the supplied code, will be monkey patched
    # to be no-ops, with a warning.
    _io_functions = ['output_notebook', 'output_file', 'show', 'save', 'reset_output']

    _loggers: dict[str, Callable[..., None]]

    _logger_text: ClassVar[str]

    _origin: ClassVar[str]

    def __init__(self, *, source: str, filename: PathLike, argv: list[str] = [], package: ModuleType | None = None) -> None:
        '''

        Args:
            source (str) : python source code

            filename (str) : a filename to use in any debugging or error output

            argv (list[str], optional) : a list of string arguments to make
                available as ``sys.argv`` when the code executes

        '''
        super().__init__()

        self._runner = CodeRunner(source, filename, argv, package=package)

        self._loggers = {}
        for f in CodeHandler._io_functions:
            self._loggers[f] = self._make_io_logger(f)

    # Properties --------------------------------------------------------------

    @property
    def error(self) -> str | None:
        ''' If the handler fails, may contain a related error message.

        '''
        return self._runner.error

    @property
    def error_detail(self) -> str | None:
        ''' If the handler fails, may contain a traceback or other details.

        '''
        return self._runner.error_detail

    @property
    def failed(self) -> bool:
        ''' ``True`` if the handler failed to modify the doc

        '''
        return self._runner.failed

    @property
    def safe_to_fork(self) -> bool:
        ''' Whether it is still safe for the Bokeh server to fork new workers.

        ``False`` if the code has already been executed.

        '''
        return not self._runner.ran

    # Public methods ----------------------------------------------------------

    def modify_document(self, doc: Document) -> None:
        ''' Run Bokeh application code to update a ``Document``

        Args:
            doc (Document) : a ``Document`` to update

        '''

        module = self._runner.new_module()

        # If no module was returned it means the code runner has some permanent
        # unfixable problem, e.g. the configured source code has a syntax error
        if module is None:
            return

        # One reason modules are stored is to prevent the module from being gc'd
        # before the document is. A symptom of a gc'd module is that its globals
        # become None. Additionally stored modules are used to provide correct
        # paths to custom models resolver.
        doc.modules.add(module)

        with _monkeypatch_io(self._loggers):
            with patch_curdoc(doc):
                self._runner.run(module, self._make_post_doc_check(doc))

    def url_path(self) -> str | None:
        ''' The last path component for the basename of the configured filename.

        '''
        if self.failed:
            return None

        # TODO should fix invalid URL characters
        return '/' + splitext(basename(self._runner.path))[0]

    # Private methods ---------------------------------------------------------

    # subclasses must define self._logger_text
    def _make_io_logger(self, name: str) -> Callable[..., None]:
        def logger(*args: Any, **kwargs: Any) -> None:
            log.info(self._logger_text, self._runner.path, name)
        return logger

    # script is supposed to edit the doc not replace it
    def _make_post_doc_check(self, doc: Document) -> Callable[[], None]:
        def func() -> None:
            if curdoc() is not doc:
                raise RuntimeError(f"{self._origin} at {self._runner.path!r} replaced the output document")
        return func

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

# monkeypatching is a little ugly, but in this case there's no reason any legitimate
# code should be calling these functions, and we're only making a best effort to
# warn people so no big deal if we fail.
@contextmanager
def _monkeypatch_io(loggers: dict[str, Callable[..., None]]) -> dict[str, Any]:
    import bokeh.io as io
    old: dict[str, Any] = {}
    for f in CodeHandler._io_functions:
        old[f] = getattr(io, f)
        setattr(io, f, loggers[f])
    yield
    for f in old:
        setattr(io, f, old[f])

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Any

# Bokeh imports
from ..core.has_props import abstract
from ..model import Model

@abstract
@dataclass(init=False)
class LabelingPolicy(Model):
    ...

@dataclass
class AllLabels(LabelingPolicy):
    ...

@dataclass
class NoOverlap(LabelingPolicy):

    min_distance: int = ...

@dataclass
class CustomLabelingPolicy(LabelingPolicy):

    args: dict[str, Any] = ...

    code: str = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..core.has_props import abstract
from ..model import Model

@abstract
@dataclass(init=False)
class BaseText(Model):

    text: str = ...

@abstract
@dataclass(init=False)
class MathText(BaseText):
    ...

@dataclass
class Ascii(MathText):
    ...

@dataclass
class MathML(MathText):
    ...

@dataclass
class TeX(MathText):

    macros: dict[str, str | tuple[str, int]] = ...

    inline: bool = ...

@dataclass
class PlainText(BaseText):
    ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Literal

# Bokeh imports
from ...core.enums import MovableType as Movable, ResizableType as Resizable
from ..dom import DOMNode
from ..nodes import Node
from .ui_element import UIElement

@dataclass
class Dialog(UIElement):

    title: str | DOMNode | UIElement | None = ...

    content: str | DOMNode | UIElement = ...

    pinnable: bool = ...

    collapsible: bool = ...

    minimizable: bool = ...

    maximizable: bool = ...

    closable: bool = ...

    close_action: Literal["hide", "destroy"] = ...

    resizable: Resizable = ...

    movable: Movable = ...

    symmetric: bool = ...

    top_limit: Node | None = ...

    bottom_limit: Node | None = ...

    left_limit: Node | None = ...

    right_limit: Node | None = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
'''

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import Any

# Bokeh imports
from ...core.has_props import abstract
from ...core.properties import (
    Bool,
    Enum,
    Int,
    List,
    Nullable,
    String,
)
from .buttons import ButtonLike
from .widget import Widget

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'AbstractGroup',
    'CheckboxButtonGroup',
    'CheckboxGroup',
    'RadioButtonGroup',
    'RadioGroup',
    'ToggleButtonGroup',
    'ToggleInputGroup',
)

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

@abstract
class AbstractGroup(Widget):
    ''' Abstract base class for all kinds of groups.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    labels = List(String, help="""
    List of text labels contained in this group.
    """)

@abstract
class ToggleButtonGroup(AbstractGroup, ButtonLike):
    ''' Abstract base class for groups with items rendered as buttons.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    orientation = Enum("horizontal", "vertical", help="""
    Orient the button group either horizontally (default) or vertically.
    """)

@abstract
class ToggleInputGroup(AbstractGroup):
    ''' Abstract base class for groups with items rendered as check/radio
    boxes.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    inline = Bool(False, help="""
    Should items be arrange vertically (``False``) or horizontally
    in-line (``True``).
    """)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

class CheckboxGroup(ToggleInputGroup):
    ''' A group of check boxes.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    active = List(Int, help="""
    The list of indices of selected check boxes.
    """)

class RadioGroup(ToggleInputGroup):
    ''' A group of radio boxes.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    active = Nullable(Int, help="""
    The index of the selected radio box, or ``None`` if nothing is selected.
    """)

class CheckboxButtonGroup(ToggleButtonGroup):
    ''' A group of check boxes rendered as toggle buttons.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    active = List(Int, help="""
    The list of indices of selected check boxes.
    """)

class RadioButtonGroup(ToggleButtonGroup):
    ''' A group of radio boxes rendered as toggle buttons.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    active = Nullable(Int, help="""
    The index of the selected radio box, or ``None`` if nothing is
    selected.
    """)

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Models for computing good tick locations on different kinds
of plots.

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import Any

# Bokeh imports
from ..core.enums import LatLon
from ..core.has_props import abstract
from ..core.properties import (
    AnyRef,
    Auto,
    Dict,
    Either,
    Enum,
    Float,
    Instance,
    Int,
    NonEmpty,
    Nullable,
    Override,
    Required,
    Seq,
    String,
)
from ..core.validation import error
from ..core.validation.errors import MISSING_MERCATOR_DIMENSION
from ..model import Model
from .mappers import ScanningColorMapper

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'AdaptiveTicker',
    'BasicTicker',
    'BinnedTicker',
    'CategoricalTicker',
    'CompositeTicker',
    'ContinuousTicker',
    'CustomJSTicker',
    'DatetimeTicker',
    'DaysTicker',
    'FixedTicker',
    'LogTicker',
    'MercatorTicker',
    'MonthsTicker',
    'SingleIntervalTicker',
    'Ticker',
    'YearsTicker',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

@abstract
class Ticker(Model):
    ''' A base class for all ticker types.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class CustomJSTicker(Ticker):
    ''' Generate tick locations that are computed by a user-defined function.

    A ``CustomJSTicker`` may be used with either a continuous (numeric) axis,
    or a categorical axis. However, only basic, non-hierarchical categorical
    axes (i.e. with a single level of factors) are supported.

    .. warning::
        The explicit purpose of this Bokeh Model is to embed *raw JavaScript
        code* for a browser to execute. If any part of the code is derived
        from untrusted user inputs, then you must take appropriate care to
        sanitize the user input prior to passing to Bokeh.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    args = Dict(String, AnyRef, help="""
    A mapping of names to Python objects. In particular those can be bokeh's models.
    These objects are made available to the ticker's code snippet as the values of
    named parameters to the callback.
    """)

    major_code = String(default="", help="""
    Callback code to run in the browser to compute minor tick locations for the
    current viewport.

    The ``cb_data`` parameter that is available to the callback code will contain
    four specific fields:

    ``start``
        the computed start coordinate of the axis

    ``end``
        the computed end of the axis

    ``range``
        the Range model for this axis

    ``cross_loc``
        the coordinate that this axis intersects the orthogonal axis
    """)

    minor_code = String(default="", help="""
    Callback code to run in the browser to compute minor tick locations for the
    current viewport.

    .. note::
        Minor ticks are not used for categorical axes. This property will be
        ignored when the range is a ``FactorRange``.

    The ``cb_data`` parameter that is available to the callback code will contain
    five specific fields:

    ``major_ticks``
        the list of the current computed major tick locations

    ``start``
        the computed start coordinate of the axis

    ``end``
        the computed end of the axis

    ``range``
        the Range model for this axis

    ``cross_loc``
        the coordinate that this axis intersects the orthogonal axis
    """)

@abstract
class ContinuousTicker(Ticker):
    ''' A base class for non-categorical ticker types.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    num_minor_ticks = Int(5, help="""
    The number of minor tick positions to generate between
    adjacent major tick values.
    """)

    desired_num_ticks = Int(6, help="""
    A desired target number of major tick positions to generate across
    the plot range.

    .. note:
        This value is a suggestion, and ticker subclasses may ignore
        it entirely, or use it only as an ideal goal to approach as well
        as can be, in the context of a specific ticking strategy.
    """)

class FixedTicker(ContinuousTicker):
    ''' Generate ticks at fixed, explicitly supplied locations.

    .. note::
        The ``desired_num_ticks`` property is ignored by this Ticker.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    ticks = Seq(Float, default=[], help="""
    List of major tick locations.
    """)

    minor_ticks = Seq(Float, default=[], help="""
    List of minor tick locations.
    """)

class AdaptiveTicker(ContinuousTicker):
    ''' Generate "nice" round ticks at any magnitude.

    Creates ticks that are "base" multiples of a set of given
    mantissas. For example, with ``base=10`` and
    ``mantissas=[1, 2, 5]``, the ticker will generate the sequence::

        ..., 0.1, 0.2, 0.5, 1, 2, 5, 10, 20, 50, 100, ...

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    base = Float(10.0, help="""
    The multiplier to use for scaling mantissas.
    """)

    mantissas = Seq(Float, default=[1, 2, 5], help="""
    The acceptable list numbers to generate multiples of.
    """)

    min_interval = Float(0.0, help="""
    The smallest allowable interval between two adjacent ticks.
    """)

    max_interval = Nullable(Float, help="""
    The largest allowable interval between two adjacent ticks.

    .. note::
        To specify an unbounded interval, set to ``None``.
    """)

class CompositeTicker(ContinuousTicker):
    ''' Combine different tickers at different scales.

    Uses the ``min_interval`` and ``max_interval`` interval attributes
    of the tickers to select the appropriate ticker at different
    scales.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    tickers = NonEmpty(Seq(Instance(Ticker)), help="""
    A list of ``Ticker`` objects to combine at different scales in order
    to generate tick values. The supplied tickers should be in order.
    Specifically, if S comes before T, then it should be the case that:

    .. code-block:: javascript

        S.get_max_interval() < T.get_min_interval()

    """)

class BaseSingleIntervalTicker(ContinuousTicker):
    ''' Base class for single interval tickers. '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class SingleIntervalTicker(BaseSingleIntervalTicker):
    ''' Generate evenly spaced ticks at a fixed interval regardless of
    scale.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    interval = Required(Float, help="""
    The interval between adjacent ticks.
    """)

class DaysTicker(BaseSingleIntervalTicker):
    ''' Generate ticks spaced apart by specific, even multiples of days.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    days = Seq(Int, default=[], help="""
    The intervals of days to use.
    """)

    num_minor_ticks = Override(default=0)

class MonthsTicker(BaseSingleIntervalTicker):
    ''' Generate ticks spaced apart by specific, even multiples of months.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    months = Seq(Int, default=[], help="""
    The intervals of months to use.
    """)

class YearsTicker(BaseSingleIntervalTicker):
    ''' Generate ticks spaced apart even numbers of years.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class BasicTicker(AdaptiveTicker):
    ''' Generate ticks on a linear scale.

    .. note::
        This class may be renamed to ``LinearTicker`` in the future.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class LogTicker(AdaptiveTicker):
    ''' Generate ticks on a log scale.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    mantissas = Override(default=[1, 5])


class MercatorTicker(BasicTicker):
    ''' Generate nice lat/lon ticks form underlying WebMercator coordinates.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    dimension = Nullable(Enum(LatLon), help="""
    Specify whether to generate ticks for Latitude or Longitude.

    Projected coordinates are not separable, computing Latitude and Longitude
    tick locations from Web Mercator requires considering coordinates from
    both dimensions together. Use this property to specify which result should
    be returned.

    Typically, if the ticker is for an x-axis, then dimension should be
    ``"lon"`` and if the ticker is for a y-axis, then the dimension
    should be `"lat"``.

    In order to prevent hard to debug errors, there is no default value for
    dimension. Using an un-configured ``MercatorTicker`` will result in a
    validation error and a JavaScript console error.
    """)

    @error(MISSING_MERCATOR_DIMENSION)
    def _check_missing_dimension(self):
        if self.dimension is None:
            return str(self)

class CategoricalTicker(Ticker):
    ''' Generate ticks for categorical ranges.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

ONE_MILLI = 1.0
ONE_SECOND = 1000.0
ONE_MINUTE = 60.0 * ONE_SECOND
ONE_HOUR = 60 * ONE_MINUTE
ONE_DAY = 24 * ONE_HOUR
ONE_MONTH = 30 * ONE_DAY # An approximation, obviously.
ONE_YEAR = 365 * ONE_DAY

class DatetimeTicker(CompositeTicker):
    ''' Generate nice ticks across different date and time scales.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    num_minor_ticks = Override(default=0)

    # TODO: (bev) InstanceDefault for this, someday
    tickers = Override(default=lambda: [
        AdaptiveTicker(
            mantissas=[1, 2, 5],
            base=10,
            min_interval=0,
            max_interval=500*ONE_MILLI,
            num_minor_ticks=0,
        ),
        AdaptiveTicker(
            mantissas=[1, 2, 5, 10, 15, 20, 30],
            base=60,
            min_interval=ONE_SECOND,
            max_interval=30*ONE_MINUTE,
            num_minor_ticks=0,
        ),
        AdaptiveTicker(
            mantissas=[1, 2, 4, 6, 8, 12],
            base=24,
            min_interval=ONE_HOUR,
            max_interval=12*ONE_HOUR,
            num_minor_ticks=0,
        ),
        DaysTicker(days=list(range(1, 32))),
        DaysTicker(days=list(range(1, 31, 3))),
        DaysTicker(days=[1, 8, 15, 22]),
        DaysTicker(days=[1, 15]),

        MonthsTicker(months=list(range(0, 12, 1))),
        MonthsTicker(months=list(range(0, 12, 2))),
        MonthsTicker(months=list(range(0, 12, 4))),
        MonthsTicker(months=list(range(0, 12, 6))),

        YearsTicker(),
    ])

class BinnedTicker(Ticker):
    """ Ticker that aligns ticks exactly at bin boundaries of a scanning color mapper.

    """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    mapper = Instance(ScanningColorMapper, help="""
    A scanning color mapper (e.g. ``EqHistColorMapper``) to use.
    """)

    num_major_ticks = Either(Int, Auto, default=8, help="""
    The number of major tick positions to show or "auto" to use the
    number of bins provided by the mapper.
    """)

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from datetime import datetime as DateTime, timedelta as TimeDelta
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Sequence,
    TypeAlias,
    TypeVar,
)

# External imports
import numpy as np
import numpy.typing as npt

if TYPE_CHECKING:
    import PIL.Image

# Bokeh imports
from .. import enums
from ..enums import AutoType as Auto
from .bases import Property

DashPatternType: TypeAlias = enums.DashPatternType | str | Sequence[int]
DashPattern: TypeAlias = Property[DashPatternType]

ImageType: TypeAlias = str | Path | PIL.Image.Image | npt.NDArray[np.uint8]
Image: TypeAlias = Property[ImageType]

HatchPatternTypeType: TypeAlias = enums.HatchPatternType | enums.HatchPatternAbbreviationType
HatchPatternType: TypeAlias = Property[HatchPatternTypeType]

T = TypeVar("T")

Bounds: TypeAlias = tuple[T, T] | tuple[T | None, T] | tuple[T, T | None]

MinMaxBoundsType: TypeAlias = Auto | Bounds[float] | Bounds[DateTime] | Bounds[TimeDelta]
MinMaxBounds: TypeAlias = Property[MinMaxBoundsType]

CSSLengthType: TypeAlias = str
CSSLength: TypeAlias = Property[CSSLengthType]

FontSizeType: TypeAlias = str
FontSize: TypeAlias = Property[FontSizeType]

MarkerTypeType: TypeAlias = enums.MarkerTypeType
MarkerType: TypeAlias = Property[MarkerTypeType]



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...core.has_props import abstract
from ...model import Model

@abstract
@dataclass(init=False)
class GroupBy(Model):
    ...

@dataclass
class GroupByModels(GroupBy):

    groups: list[list[Model]] = ...

@dataclass
class GroupByName(GroupBy):
    ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import InitVar, dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Sequence,
    TypeAlias,
)

# External imports
import numpy as np
import numpy.typing as npt

if TYPE_CHECKING: # to work around ruff(TID253)
    from narwhals.stable.v1.typing import IntoDataFrame, IntoSeries
    from pandas import DataFrame
    from pandas.core.groupby import GroupBy

# Bokeh imports
from .._types import (
    Color,
    Datetime,
    TextLike,
    TimeDelta,
)
from ..core.enums import (
    AutoType as Auto,
    HexTileOrientationType as HexTileOrientation,
    HorizontalLocationType as HorizontalLocation,
    PaletteType as Palette,
    VerticalLocationType as VerticalLocation,
)
from ..models.dom import Template
from ..models.glyphs import (
    HArea,
    HBar,
    HexTile,
    Line,
    VArea,
    VBar,
)
from ..models.graphs import LayoutProvider
from ..models.plots import Plot
from ..models.ranges import Range
from ..models.renderers import ContourRenderer, GlyphRenderer, GraphRenderer
from ..models.scales import Scale
from ..models.sources import ColumnDataSource
from ..models.tools import (
    Drag,
    GestureTool,
    InspectTool,
    Scroll,
    Tap,
    Tool,
)
from .glyph_api import GlyphAPI

EagerDataFrame: TypeAlias = IntoDataFrame
EagerSeries: TypeAlias = IntoSeries

@dataclass(init=False)
class BaseFigureOptions:

    tools: InitVar[str | Sequence[str | Tool]] = ...

    x_minor_ticks: InitVar[Auto | int] = ...

    y_minor_ticks: InitVar[Auto | int] = ...

    x_axis_location: InitVar[VerticalLocation | None] = ...

    y_axis_location: InitVar[HorizontalLocation | None] = ...

    x_axis_label: InitVar[TextLike | None] = ...

    y_axis_label: InitVar[TextLike | None] = ...

    active_drag: InitVar[Auto | str | Drag | None] = ...

    active_inspect: InitVar[Auto | str | InspectTool | Sequence[InspectTool] | None] = ...

    active_scroll: InitVar[Auto | str | Scroll | None] = ...

    active_tap: InitVar[Auto | str | Tap | None] = ...

    active_multi: InitVar[Auto | str | GestureTool | None] = ...

    tooltips: InitVar[Template | str | list[tuple[str, str]] | None] = ...

RangeLike: TypeAlias = (
    Range |
    tuple[float, float] |
    tuple[Datetime, Datetime] |
    tuple[TimeDelta, TimeDelta] |
    Sequence[str] |
    EagerSeries |
    GroupBy[Any]
)

AxisType: TypeAlias = Auto | Literal["linear", "log", "datetime", "mercator"] | None

@dataclass(init=False)
class FigureOptions(BaseFigureOptions):

    x_range: InitVar[RangeLike] = ...

    y_range: InitVar[RangeLike] = ...

    x_axis_type: InitVar[AxisType] = ...

    y_axis_type: InitVar[AxisType] = ...

@dataclass
class figure(Plot, GlyphAPI, FigureOptions):

    def subplot(self,
        *,
        x_source: Range | None = ...,
        y_source: Range | None = ...,
        x_scale: Scale | None = ...,
        y_scale: Scale | None = ...,
        x_target: Range,
        y_target: Range,
    ) -> GlyphAPI: ...

    def hexbin(self,
        x: npt.NDArray[np.floating],
        y: npt.NDArray[np.floating],
        size: float,
        orientation: HexTileOrientation = "pointytop",
        palette: Palette = "Viridis256",
        line_color: Color | None = None,
        fill_color: Color | None = None,
        aspect_scale: float = 1,
        **kwargs: Any, # TODO Unpack[HexTileArgs]
    ) -> tuple[GlyphRenderer[HexTile], DataFrame]: ...

    def harea_stack(self, stackers: Sequence[str], **kwargs: Any) -> list[GlyphRenderer[HArea]]: ...

    def varea_stack(self, stackers: Sequence[str], **kwargs: Any) -> list[GlyphRenderer[VArea]]: ...

    def hbar_stack(self, stackers: Sequence[str], **kwargs: Any) -> list[GlyphRenderer[HBar]]: ...

    def vbar_stack(self, stackers: Sequence[str], **kwargs: Any) -> list[GlyphRenderer[VBar]]: ...

    def hline_stack(self, stackers: Sequence[str], **kwargs: Any) -> list[GlyphRenderer[Line]]: ...

    def vline_stack(self, stackers: Sequence[str], **kwargs: Any) -> list[GlyphRenderer[Line]]: ...

    def graph(self, node_source: ColumnDataSource, edge_source: ColumnDataSource, layout_provider: LayoutProvider, **kwargs: Any) -> GraphRenderer: ...

    def contour(self,
        x: npt.ArrayLike | None = None,
        y: npt.ArrayLike | None = None,
        z: npt.ArrayLike | np.ma.MaskedArray[Any, Any] | None = None,
        levels: npt.ArrayLike | None = None,
        **visuals: Any,
    ) -> ContourRenderer: ...

def markers() -> None: ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..core.enums import TextureRepetitionType as TextureRepetition
from ..core.has_props import abstract
from ..core.property.visual import ImageType as Image
from ..model import Model

@abstract
@dataclass(init=False)
class Texture(Model):

    repetition: TextureRepetition = ...

@dataclass
class CanvasTexture(Texture):

    code: str = ...

@dataclass
class ImageURLTexture(Texture):

    url: Image = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import Any

# Bokeh imports
from ..core.has_props import abstract
from ..core.properties import (
    AnyRef,
    Bool,
    Instance,
    Int,
    NonEmpty,
    Nullable,
    Required,
    RestrictedDict,
    Seq,
    String,
)
from ..model import Model

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    "AllIndices",
    "BooleanFilter",
    "CustomJSFilter",
    "DifferenceFilter",
    "Filter",
    "GroupFilter",
    "IndexFilter",
    "IntersectionFilter",
    "InversionFilter",
    "SymmetricDifferenceFilter",
    "UnionFilter",
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

@abstract
class Filter(Model):
    ''' A Filter model represents a filtering operation that returns a row-wise subset of
    data when applied to a ``ColumnDataSource``.
    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def __invert__(self) -> Filter:
        return InversionFilter(operand=self)

    def __and__(self, other: Filter) -> Filter:
        return IntersectionFilter(operands=[self, other])

    def __or__(self, other: Filter) -> Filter:
        return UnionFilter(operands=[self, other])

    def __sub__(self, other: Filter) -> Filter:
        return DifferenceFilter(operands=[self, other])

    def __xor__(self, other: Filter) -> Filter:
        return SymmetricDifferenceFilter(operands=[self, other])

class AllIndices(Filter):
    """ Trivial filter that includes all indices in a dataset. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class InversionFilter(Filter):
    """ Inverts indices resulting from another filter. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    operand = Required(Instance(Filter), help="""
    Indices produced by this filter will be inverted.
    """)

@abstract
class CompositeFilter(Filter):
    """ Base class for composite filters. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    operands = Required(NonEmpty(Seq(Instance(Filter))), help="""
    A collection of filters to perform an operation on.
    """)

class IntersectionFilter(CompositeFilter):
    """ Computes intersection of indices resulting from other filters. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class UnionFilter(CompositeFilter):
    """ Computes union of indices resulting from other filters. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class DifferenceFilter(CompositeFilter):
    """ Computes difference of indices resulting from other filters. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class SymmetricDifferenceFilter(CompositeFilter):
    """ Computes symmetric difference of indices resulting from other filters. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

class IndexFilter(Filter):
    ''' An ``IndexFilter`` filters data by returning the subset of data at a given set of indices.
    '''

    indices = Nullable(Seq(Int), help="""
    A list of integer indices representing the subset of data to select.
    """)

    def __init__(self, *args, **kwargs) -> None:
        if len(args) == 1 and "indices" not in kwargs:
            kwargs["indices"] = args[0]

        super().__init__(**kwargs)

class BooleanFilter(Filter):
    ''' A ``BooleanFilter`` filters data by returning the subset of data corresponding to indices
    where the values of the booleans array is True.
    '''

    booleans = Nullable(Seq(Bool), help="""
    A list of booleans indicating which rows of data to select.
    """)

    def __init__(self, *args, **kwargs) -> None:
        if len(args) == 1 and "booleans" not in kwargs:
            kwargs["booleans"] = args[0]

        super().__init__(**kwargs)

class GroupFilter(Filter):
    ''' A ``GroupFilter`` represents the rows of a ``ColumnDataSource`` where the
    values of the column indicated by ``column_name`` match the ``group`` variable.
    '''

    column_name = Required(String, help="""
    The name of the column to perform the group filtering operation on.
    """)

    group = Required(AnyRef, help="""
    The value of the column indicating the rows of data to keep.
    """)

    def __init__(self, *args, **kwargs) -> None:
        if len(args) == 2 and "column_name" not in kwargs and "group" not in kwargs:
            kwargs["column_name"] = args[0]
            kwargs["group"] = args[1]

        super().__init__(**kwargs)

class CustomJSFilter(Filter):
    ''' Filter data sources with a custom defined JavaScript function.

    .. warning::
        The explicit purpose of this Bokeh Model is to embed *raw JavaScript
        code* for a browser to execute. If any part of the code is derived
        from untrusted user inputs, then you must take appropriate care to
        sanitize the user input prior to passing to Bokeh.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    args = RestrictedDict(String, AnyRef, disallow=("source",), help="""
    A mapping of names to Python objects. In particular those can be bokeh's models.
    These objects are made available to the callback's code snippet as the values of
    named parameters to the callback.
    """)

    code = String(default="", help="""
    A snippet of JavaScript code to filter data contained in a columnar data source.
    The code is made into the body of a function, and all of of the named objects in
    ``args`` are available as parameters that the code can use. The variable
    ``source`` will contain the data source that is associated with the ``CDSView`` this
    filter is added to.

    The code should either return the indices of the subset or an array of booleans
    to use to subset data source rows.

    Example:

    .. code-block::

        code = '''
        const indices = []
        for (let i = 0; i <= source.data['some_column'].length; i++) {
            if (source.data['some_column'][i] == 'some_value') {
                indices.push(i)
            }
        }
        return indices
        '''

    """)

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Literal

# Bokeh imports
from ..core.enums import AutoType as Auto
from ..core.property_mixins import (
    ScalarBandFillProps,
    ScalarBandHatchProps,
    ScalarGridLineProps,
    ScalarMinorGridLineProps,
)
from .axes import Axis
from .renderers import GuideRenderer
from .tickers import Ticker

@dataclass
class Grid(GuideRenderer, ScalarGridLineProps, ScalarMinorGridLineProps, ScalarBandFillProps, ScalarBandHatchProps):

    dimension: Literal[0, 1] = ...

    bounds: Auto | tuple[float, float] = ...

    cross_bounds: Auto | tuple[float, float] = ...

    axis: Axis | None = ...

    ticker: Ticker | None = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Any

# Bokeh imports
from ..core.enums import AutoType as Auto
from ..core.has_props import HasProps, abstract
from ..core.types import PathLike
from ..model import Model
from ..models.ui import Dialog, UIElement

@abstract
@dataclass(init=False)
class Callback(Model):
    ...

@dataclass
class OpenURL(Callback):

    url: str = ...

    same_tab: bool = ...

@abstract
@dataclass(init=False)
class CustomCode(Callback):
    ...

@dataclass
class CustomJS(CustomCode):

    args: dict[str, Any] = ...

    code: str = ...

    module: Auto | bool = ...

    @classmethod
    def from_file(cls, path: PathLike, **args: Any) -> CustomJS: ...

@dataclass
class SetValue(Callback):

    obj: HasProps = ...

    attr: str = ...

    value: Any = ...

@dataclass
class ToggleVisibility(Callback):

    target: UIElement = ...

@dataclass
class OpenDialog(Callback):

    dialog: Dialog = ...

@dataclass
class CloseDialog(Callback):

    dialog: Dialog = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Any, Literal

# Bokeh imports
from ..._specs import ColorSpec, FontStyleSpec, TextAlignSpec
from ...core.enums import (
    AutosizeModeType as AutosizeMode,
    DateFormatType as DateFormat,
    NumeralLanguageType as NumeralLanguage,
    RoundingFunctionType as RoundingFunction,
)
from ...core.has_props import abstract
from ...model import Model
from ..comparisons import Comparison
from ..sources import (
    CDSView,
    ColumnDataSource,
    DataDictLike,
    DataSource,
)
from .widget import Widget

@abstract
@dataclass(init=False)
class CellFormatter(Model):
    ...

@abstract
@dataclass(init=False)
class CellEditor(Model):
    ...

@abstract
@dataclass(init=False)
class RowAggregator(Model):

    field_: str = ...

@dataclass
class StringFormatter(CellFormatter):

    font_style: FontStyleSpec = ...

    text_align: TextAlignSpec = ...

    text_color: ColorSpec = ...

    background_color: ColorSpec = ...

    nan_format: str = ...

    null_format: str = ...

@dataclass
class ScientificFormatter(StringFormatter):

    precision: int = ...

    power_limit_high: int = ...

    power_limit_low: int = ...

@dataclass
class NumberFormatter(StringFormatter):

    format: str = ...

    language: NumeralLanguage = ...

    rounding: RoundingFunction = ...

@dataclass
class BooleanFormatter(CellFormatter):

    icon: Literal["check", "check-circle", "check-circle-o", "check-square", "check-square-o"] = ...

@dataclass
class DateFormatter(StringFormatter):

    format: DateFormat | str = ...

@dataclass
class HTMLTemplateFormatter(CellFormatter):

    template: str = ...

@dataclass
class StringEditor(CellEditor):

    completions: list[str] = ...

@dataclass
class TextEditor(CellEditor):
    ...

@dataclass
class SelectEditor(CellEditor):

    options: list[str] = ...

@dataclass
class PercentEditor(CellEditor):
    ...

@dataclass
class CheckboxEditor(CellEditor):
    ...

@dataclass
class IntEditor(CellEditor):

    step: int = ...

@dataclass
class NumberEditor(CellEditor):

    step: float = ...

@dataclass
class TimeEditor(CellEditor):
    ...

@dataclass
class DateEditor(CellEditor):
    ...

@dataclass
class AvgAggregator(RowAggregator):
    ...

@dataclass
class MinAggregator(RowAggregator):
    ...

@dataclass
class MaxAggregator(RowAggregator):
    ...

@dataclass
class SumAggregator(RowAggregator):
    ...

@dataclass
class TableColumn(Model):

    field: str = ...

    title: str | None = ...

    width: int = ...

    formatter: CellFormatter = ...

    editor: CellEditor = ...

    sortable: bool = ...

    default_sort: Literal["ascending", "descending"] = ...

    visible: bool = ...

    sorter: Comparison | None = ...

@abstract
@dataclass(init=False)
class TableWidget(Widget):

    source: DataSource = ...

    view: CDSView = ...

@dataclass
class DataTable(TableWidget):

    autosize_mode: AutosizeMode = ...

    auto_edit: bool = ...

    columns: list[TableColumn] = ...

    fit_columns: bool | None = ...

    frozen_columns: int | None = ...

    frozen_rows: int | None = ...

    sortable: bool = ...

    reorderable: bool = ...

    editable: bool = ...

    selectable: bool | Literal["checkbox"] = ...

    index_position: int | None = ...

    index_header: str = ...

    index_width: int = ...

    scroll_to_selection: bool = ...

    header_row: bool = ...

    row_height: int = ...

    @staticmethod
    def from_data(data: ColumnDataSource | DataDictLike, columns: list[str] | None = None,
                  formatters: dict[str, CellFormatter] = {}, **kwargs: Any) -> DataTable: ...

@dataclass
class GroupingInfo(Model):

    getter: str = ...

    aggregators: list[RowAggregator] = ...

    collapsed: bool = ...

@dataclass
class DataCube(DataTable):

    grouping: list[GroupingInfo] = ...

    target: DataSource = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Provide a request handler that returns a page displaying a document.

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from urllib.parse import urlparse

# Bokeh imports
from bokeh.core.templates import AUTOLOAD_JS
from bokeh.embed.bundle import Script, bundle_for_objs_and_resources
from bokeh.embed.elements import script_for_render_items
from bokeh.embed.util import RenderItem

# Bokeh imports
from .session_handler import SessionHandler

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'AutoloadJsHandler',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

class AutoloadJsHandler(SessionHandler):
    ''' Implements a custom Tornado handler for the autoload JS chunk

    '''

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "*")
        self.set_header("Access-Control-Allow-Credentials", "true")

    async def get(self, *args, **kwargs):
        if self.request.cookies and "Origin" in self.request.headers:
            # If credentials, i.e. cookies, are sent with the request,
            # we cannot leave the allowed origin as wildcard "*",
            # but have to make it explicit.
            self.set_header("Access-Control-Allow-Origin", self.request.headers["Origin"])

        session = await self.get_session()

        element_id = self.get_argument("bokeh-autoload-element", default=None)
        if not element_id:
            self.send_error(status_code=400, reason='No bokeh-autoload-element query parameter')
            return

        app_path = self.get_argument("bokeh-app-path", default="/")
        absolute_url = self.get_argument("bokeh-absolute-url", default=None)

        if absolute_url:
            uri = urlparse(absolute_url)
            server_url = f"{uri.scheme}://{uri.netloc}"
        else:
            server_url = None

        resources_param = self.get_argument("resources", "default")
        resources = self.application.resources(server_url) if resources_param != "none" else None
        bundle = bundle_for_objs_and_resources(None, resources)

        render_items = [RenderItem(token=session.token, elementid=element_id, use_for_title=False)]
        bundle.add(Script(script_for_render_items({}, render_items, app_path=app_path, absolute_url=absolute_url)))

        js = AUTOLOAD_JS.render(bundle=bundle, elementid=element_id)

        self.set_header("Content-Type", 'application/javascript')
        self.write(js)

    async def options(self, *args, **kwargs):
        '''Browsers make OPTIONS requests under the hood before a GET request'''
        self.set_header('Access-Control-Allow-Methods', 'PUT, GET, OPTIONS')
        self.set_header("Access-Control-Allow-Origin", self.request.headers["Origin"])

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ...tools import Toolbar
from .html_annotation import HTMLAnnotation

@dataclass
class ToolbarPanel(HTMLAnnotation):

    toolbar: Toolbar = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import (
    Any,
    Literal,
    Sequence,
    TypeAlias,
)

# Bokeh imports
from ..._types import (
    Color,
    ColorHex,
    NonNegative,
    Positive,
    Readonly,
)
from ...core.enums import AutoType as Auto
from ...core.has_props import abstract
from ...core.property_aliases import IconLikeType as IconLike
from ...events import ModelEvent
from ..dom import HTML
from ..formatters import TickFormatter
from ..ui import Tooltip
from .widget import Widget

@dataclass
class ClearInput(ModelEvent):

    def __init__(self, model: InputWidget) -> None: ...

@abstract
@dataclass(init=False)
class InputWidget(Widget):

    title: str | HTML = ...

    description: str | Tooltip | None = ...

@dataclass
class FileInput(InputWidget):

    value: Readonly[str | list[str]] = ...

    mime_type: Readonly[str | list[str]] = ...

    filename: Readonly[str | list[str]] = ...

    accept: str | list[str] = ...

    multiple: bool = ...

    directory: bool = ...

    def clear(self) -> None: ...

@dataclass
class NumericInput(InputWidget):

    value: None | float | int = ...

    low: None | float | int = ...

    high: None | float | int = ...

    placeholder: str = ...

    mode: Literal["int", "float"] = ...

    format: None | str | TickFormatter = ...

@dataclass
class Spinner(NumericInput):

    value_throttled: Readonly[int | float | None] = ...

    step: float = ...

    page_step_multiplier: float = ...

    wheel_wait: int | float = ...

@abstract
@dataclass(init=False)
class ToggleInput(Widget):

    active: bool = ...

    label: str = ...

@dataclass
class Checkbox(ToggleInput):
    ...

@dataclass
class Switch(ToggleInput):

    on_icon: IconLike | None = ...

    off_icon: IconLike | None = ...

@dataclass
class TextLikeInput(InputWidget):

    value: str = ...

    value_input: str = ...

    placeholder: str = ...

    max_length: int | None = ...

@dataclass
class TextInput(TextLikeInput):

    prefix: str | None = ...

    suffix: str | None = ...

@dataclass
class TextAreaInput(TextLikeInput):

    cols: int = ...

    rows: int = ...

@dataclass
class PasswordInput(TextInput):
    ...

@dataclass
class AutocompleteInput(TextInput):

    completions: list[str] = ...

    max_completions: Positive[int] | None = ...

    min_characters: NonNegative[int] = ...

    case_sensitive: bool = ...

    restrict: bool = ...

    search_strategy: Literal["starts_with", "includes"] = ...

Options: TypeAlias = list[str | tuple[Any, str]]
OptionsGroups: TypeAlias = dict[str, Options]

@dataclass
class Select(InputWidget):

    options: Options | OptionsGroups | list[str | None] = ...

    value: Any = ...

@dataclass
class MultiSelect(InputWidget):

    options: list[str | tuple[str, str]] = ...

    value: list[str] = ...

    size: int = ...

@dataclass
class MultiChoice(InputWidget):

    options: list[str | tuple[str, str]] = ...

    value: list[str] = ...

    delete_button: bool = ...

    max_items: int | None = ...

    option_limit: int | None = ...

    search_option_limit: int | None = ...

    placeholder: str | None = ...

    solid: bool = ...

@dataclass
class ColorPicker(InputWidget):

    color: ColorHex = ...

@dataclass
class PaletteSelect(InputWidget):

    value: str = ...

    items: Sequence[tuple[str, Sequence[Color]]] = ...

    swatch_width: NonNegative[int] = ...

    swatch_height: Auto | NonNegative[int] = ...

    ncols: Positive[int] = ...

def ColorMap(*args: Any, **kwargs: Any) -> PaletteSelect: ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
import datetime
from typing import TypeAlias, TypeVar

# Bokeh imports
from .models.nodes import Node
from .models.ranges import Factor
from .models.text import BaseText

T = TypeVar("T")

NonNegative: TypeAlias = T
Positive: TypeAlias = T
Readonly: TypeAlias = T

TextLike: TypeAlias = str | BaseText

Date: TypeAlias = str | datetime.date
Datetime: TypeAlias = str | datetime.date | datetime.datetime
Time: TypeAlias = str | datetime.time
TimeDelta: TypeAlias = datetime.timedelta

Color: TypeAlias = str | tuple[int, int, int] | tuple[int, int, int, float]
ColorHex: TypeAlias = Color

Alpha: TypeAlias = float
Size: TypeAlias = float
Angle: TypeAlias = float
Percent: TypeAlias = float

FontSize: TypeAlias = str
FontStyle: TypeAlias = str

Regex: TypeAlias = str

Bytes: TypeAlias = bytes
JSON: TypeAlias = str

CoordinateLike: TypeAlias = float | Datetime | Factor
Coordinate: TypeAlias = CoordinateLike | Node



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
from __future__ import annotations

# Standard library imports
import os, re, subprocess, sys, time
from itertools import product
from pathlib import Path
from shutil import copy, copytree, rmtree
from textwrap import indent
from typing import NoReturn

# External imports
from setuptools import setup  # type: ignore[import]
from setuptools.command.build import build  # type: ignore[import]
from setuptools.command.editable_wheel import editable_wheel  # type: ignore[import]
from setuptools.command.sdist import sdist  # type: ignore[import]

# --- Helpers ----------------------------------------------------------------

try:
    import colorama
    if sys.platform == "win32": colorama.init()
    def bright(text: str) -> str: return f"{colorama.Style.BRIGHT}{text}{colorama.Style.RESET_ALL}"
    def dim(text: str) -> str: return f"{colorama.Style.DIM}{text}{colorama.Style.RESET_ALL}"
    def red(text: str) -> str: return f"{colorama.Fore.RED}{text}{colorama.Style.RESET_ALL}"
    def green(text: str) -> str: return f"{colorama.Fore.GREEN}{text}{colorama.Style.RESET_ALL}"
    def yellow(text: str) -> str: return f"{colorama.Fore.YELLOW}{text}{colorama.Style.RESET_ALL}"
except ModuleNotFoundError:
    def _plain(text: str) -> str: return text
    bright = dim = red = green = yellow = _plain

ROOT = Path(__file__).resolve().parent
SRC_ROOT = ROOT / "src"
BUILD_JS = ROOT / 'bokehjs' / 'build' / 'js'
BUILD_TSLIB = ROOT / 'bokehjs' / 'node_modules' / 'typescript' / 'lib'
PKG_STATIC = SRC_ROOT / 'bokeh' / 'server' / 'static'
PKG_JS = PKG_STATIC / 'js'
PKG_TSLIB = PKG_STATIC / 'lib'
COMPONENTS = ("bokeh", "bokeh-widgets", "bokeh-tables", "bokeh-api", "bokeh-gl", "bokeh-mathjax")
JS_FILES = [f"{c}{m}.js" for c, m in product(COMPONENTS, ("", ".min"))]
PACKAGED = (ROOT / 'PKG-INFO').exists()

def build_js() -> None:
    print("\nBuilding BokehJS... ", end="")
    try:
        t0 = time.time()
        proc = subprocess.run(["node", "make", "build"], capture_output=True, cwd="bokehjs")
        t1 = time.time()
    except OSError as e:
        die(BUILD_EXEC_FAIL_MSG.format(exc=e))

    if proc.returncode != 0:
        out = indent(proc.stdout.decode('ascii', errors='ignore'), '    ')
        err = indent(proc.stderr.decode('ascii', errors='ignore'), '    ')
        die(BUILD_FAIL_MSG.format(stdout=red(out), stderr=red(err)))

    out = proc.stdout.decode('ascii', errors='ignore')
    pat = re.compile(r"(\[.*\]) (.*)", re.DOTALL)
    msg = []
    for line in out.strip().split("\n"):
        if m := pat.match(line):
            stamp, txt = m.groups()
            msg.append(f"   {dim(green(stamp))} {dim(txt)}")
    print(BUILD_SUCCESS_MSG.format(msg="\n".join(msg)))
    print(f"\n Build time: {bright(yellow(f'{t1-t0:0.1f} seconds'))}\n")

    print("Build artifact sizes:")
    try:
        for fn in JS_FILES:
            size = (BUILD_JS / fn).stat().st_size / 2**10
            print(f"  - {fn:<20} : {size:6.1f} KB")
    except FileNotFoundError as e:
        die(BUILD_SIZE_FAIL_MSG.format(exc=e))

def install_js(packages: list[str]) -> None:
    print("\nInstalling BokehJS... ", end="")

    missing = [fn for fn in JS_FILES if not (BUILD_JS / fn).exists()]
    if missing:
        die(BOKEHJS_INSTALL_FAIL.format(missing=", ".join(missing)))

    if not PKG_JS.is_symlink():
        if PKG_JS.exists():
            rmtree(PKG_JS)
        copytree(BUILD_JS, PKG_JS)

    if not PKG_TSLIB.is_symlink():
        if PKG_TSLIB.exists():
            rmtree(PKG_TSLIB)
        if BUILD_TSLIB.exists():
            PKG_TSLIB.mkdir()
            for lib_file in BUILD_TSLIB.glob("lib.*.d.ts"):
                copy(lib_file, PKG_TSLIB)

    new = set(
        ".".join([*Path(parent).relative_to(SRC_ROOT).parts, d])
        for parent, dirs, _ in os.walk(PKG_STATIC) for d in dirs
    )
    existing = set(packages)
    packages.extend(tuple(new-existing))

    print(SUCCESS)

def build_or_install_bokehjs(packages: list[str]) -> None:
    action = os.environ.get("BOKEHJS_ACTION", "build")
    if PACKAGED:
        kind, loc = "PACKAGED", "bokeh.server.static"
    elif action == "install":
        kind, loc = "PREVIOUSLY BUILT", "bokehjs/build"
        install_js(packages)
    elif action == "build":
        kind, loc = "NEWLY BUILT", "bokehjs/build"
        build_js()
        install_js(packages)
    else:
        raise ValueError(f"Unrecognized action {action!r}")
    print(f"Used {bright(yellow(kind))} BokehJS from {loc}\n")

def check_tags() -> None:
    if not PACKAGED:
        try:
            tags = subprocess.check_output(("git", "tag", "-l")).split()
            if len(tags) < 5: # arbitrary "too-low" cutoff, there will always be more than 5 tags, if there are any
                die(bright(red(MISSING_TAGS)))
        except Exception:
            print(bright(yellow("!!! Could not check repo tags. Please ensure full tag history")))

def die(x: str) -> NoReturn:
    print(f"{x}\n")
    sys.exit(1)

MISSING_TAGS = """
!!! This repository is missing git tags! Full tag history is required to build Bokeh.
!!!
!!! See https://docs.bokeh.org/en/latest/docs/dev_guide/setup.html#troubleshooting
"""
SUCCESS = f"{bright(green('Success!'))}\n"
FAILED = f"{bright(red('Failed.'))}\n"
BUILD_SUCCESS_MSG =f"{SUCCESS}\nBuild output:\n\n{{msg}}"
BUILD_SIZE_FAIL_MSG = f"{FAILED}\nERROR: could not determine sizes:\n\n     {{exc}}"
BOKEHJS_INSTALL_FAIL = f"{FAILED}\nERROR: Cannot install BokehJS: files missing in bokehjs/build:\n\n    {{missing}}"
BUILD_EXEC_FAIL_MSG = f"{FAILED}\nERROR: 'node make build' failed to execute:\n\n    {{exc}}"
BUILD_FAIL_MSG = f"""{FAILED}\nERROR: 'node make build' returned the following

---- on stdout:
{{stdout}}

---- on stderr:
{{stderr}}
"""

# --- Setuptools -------------------------------------------------------------

class Build(build):  # type: ignore
    def run(self) -> None:
        check_tags()
        build_or_install_bokehjs(self.distribution.packages)
        super().run()

class EditableWheel(editable_wheel):  # type: ignore
    def run(self) -> None:
        check_tags()
        build_or_install_bokehjs(self.distribution.packages)
        super().run()

class Sdist(sdist):  # type: ignore
    def run(self) -> None:
        check_tags()
        build_or_install_bokehjs(self.distribution.packages)
        super().run()

setup(cmdclass={"build": Build, "editable_wheel": EditableWheel, "sdist": Sdist})



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..glyph import XYGlyph
from ..glyphs import MultiLine, Patches
from ..graphs import GraphHitTestPolicy, LayoutProvider
from .glyph_renderer import GlyphRenderer
from .renderer import DataRenderer

@dataclass
class GraphRenderer(DataRenderer):

    layout_provider: LayoutProvider = ...

    node_renderer: GlyphRenderer[XYGlyph] = ...

    edge_renderer: GlyphRenderer[MultiLine | Patches] = ...

    selection_policy: GraphHitTestPolicy = ...

    inspection_policy: GraphHitTestPolicy = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Sequence, TypedDict

# Bokeh imports
from ..core.has_props import abstract
from ..model import Model

class ImageIndex(TypedDict):
   index: int
   i: int
   j: int
   flat_index: int

@dataclass
class Selection(Model):

    indices: Sequence[int] = ...

    line_indices: Sequence[int] = ...

    multiline_indices: dict[int, Sequence[int]] = ...

    image_indices: list[ImageIndex] = ...

@abstract
@dataclass(init=False)
class SelectionPolicy(Model):
    ...

@dataclass
class IntersectRenderers(SelectionPolicy):
    ...

@dataclass
class UnionRenderers(SelectionPolicy):
    ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..core.has_props import abstract
from ..model import Model

@abstract
@dataclass(init=False)
class Selector(Model):

    query: str = ...

@dataclass
class ByID(Selector):
    ...

@dataclass
class ByClass(Selector):
    ...

@dataclass
class ByCSS(Selector):
    ...

@dataclass
class ByXPath(Selector):
    ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Sequence

# Bokeh imports
from .._types import Color
from ..core.enums import MarkerTypeType as MarkerType, PaletteType as Palette
from ..core.has_props import abstract
from ..core.property.visual import HatchPatternType as HatchPattern
from .glyphs import Glyph
from .ranges import FactorSeq
from .renderers import GlyphRenderer
from .transforms import Transform

@abstract
@dataclass(init=False)
class Mapper(Transform):
    ...

@abstract
@dataclass(init=False)
class ColorMapper(Mapper):

    palette: Sequence[Color] | Palette = ...

    nan_color: Color = ...

@abstract
@dataclass(init=False)
class CategoricalMapper(Mapper):

    factors: FactorSeq = ...

    start: int = ...

    end: int | None = ...

@dataclass
class CategoricalColorMapper(CategoricalMapper, ColorMapper):
    ...

@dataclass
class CategoricalMarkerMapper(CategoricalMapper):

    markers: Sequence[MarkerType] = ...

    default_value: MarkerType = ...

@dataclass
class CategoricalPatternMapper(CategoricalMapper):

    patterns: Sequence[HatchPattern] = ...

    default_value: HatchPattern = ...

@abstract
@dataclass(init=False)
class ContinuousColorMapper(ColorMapper):

    domain: list[tuple[GlyphRenderer[Glyph], str | list[str]]] = ...

    low: float | None = ...

    high: float | None = ...

    low_color: Color | None = ...

    high_color: Color | None = ...

@dataclass
class LinearColorMapper(ContinuousColorMapper):
    ...

@dataclass
class LogColorMapper(ContinuousColorMapper):
    ...

@abstract
@dataclass(init=False)
class ScanningColorMapper(ContinuousColorMapper):
    ...

@dataclass
class EqHistColorMapper(ScanningColorMapper):

    bins: int = ...

    rescale_discrete_levels: bool = ...

@abstract
@dataclass(init=False)
class StackColorMapper(ColorMapper):
    ...

@dataclass
class WeightedStackColorMapper(StackColorMapper):

    alpha_mapper: ContinuousColorMapper = ...

    color_baseline: float | None = ...

    stack_labels: Sequence[str] | None = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
)

# Bokeh imports
from ...core.enums import AutoType as Auto
from ..annotations import ColorBar
from ..glyphs import Glyph
from ..graphics import Decoration, Marking
from ..sources import CDSView, DataSource
from .renderer import DataRenderer

GlyphType = TypeVar("GlyphType", bound=Glyph)

@dataclass
class GlyphRenderer(DataRenderer, Generic[GlyphType]):

    data_source: DataSource = ...

    view: CDSView = ...

    glyph: GlyphType = ...

    selection_glyph: Auto | GlyphType | None = ...

    nonselection_glyph: Auto | GlyphType | None = ...

    hover_glyph: GlyphType | None = ...

    muted_glyph: Auto | GlyphType | None = ...

    muted: bool = ...

    def add_decoration(self, marking: Marking, node: Literal["start", "middle", "end"]) -> Decoration: ...

    def construct_color_bar(self, **kwargs: Any) -> ColorBar: ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Literal, Sequence

# Bokeh imports
from .._types import Datetime, TextLike
from ..core.enums import (
    AlignType as Align,
    AutoType as Auto,
    LabelOrientationType as LabelOrientation,
)
from ..core.has_props import abstract
from ..core.property_mixins import (
    ScalarAxisLabelTextProps as AxisLabelText,
    ScalarAxisLineProps as AxisLine,
    ScalarBackgroundFillProps as BackgroundFill,
    ScalarBackgroundHatchProps as BackgroundHatch,
    ScalarGroupTextProps as GroupText,
    ScalarMajorLabelTextProps as MajorLabelText,
    ScalarMajorTickLineProps as MajorTickLine,
    ScalarMinorTickLineProps as MinorTickLine,
    ScalarSeparatorLineProps as SeparatorLine,
    ScalarSubgroupTextProps as SubgroupText,
)
from .formatters import (
    BasicTickFormatter,
    CategoricalTickFormatter,
    DatetimeTickFormatter,
    LogTickFormatter,
    MercatorTickFormatter,
    TickFormatter,
)
from .labeling import LabelingPolicy
from .ranges import Factor
from .renderers import GuideRenderer
from .tickers import (
    BasicTicker,
    CategoricalTicker,
    DatetimeTicker,
    LogTicker,
    MercatorTicker,
    Ticker,
)

@abstract
@dataclass(init=False)
class Axis(GuideRenderer, AxisLabelText, MajorLabelText, AxisLine, MajorTickLine, MinorTickLine, BackgroundFill, BackgroundHatch):

    dimension: Auto | Literal[0, 1] = ...

    face: Auto | Literal["front", "back"] = ...

    bounds: Auto | tuple[float, float] | tuple[Datetime, Datetime] = ...

    ticker: Ticker | Sequence[float] = ...

    formatter: TickFormatter = ...

    axis_label: TextLike | None = ...

    axis_label_standoff: int = ...

    axis_label_orientation: LabelOrientation | float = ...

    axis_label_align: Align = ...

    major_label_standoff: int = ...

    major_label_orientation: LabelOrientation | float = ...

    major_label_overrides: dict[float | str, TextLike] = ...

    major_label_policy: LabelingPolicy = ...

    major_tick_in: int = ...

    major_tick_out: int = ...

    minor_tick_in: int = ...

    minor_tick_out: int = ...

    fixed_location: None | float | Factor = ...

@abstract
@dataclass(init=False)
class ContinuousAxis(Axis):
    ...

@dataclass
class LinearAxis(ContinuousAxis):

    ticker: BasicTicker = ...

    formatter: BasicTickFormatter = ...

@dataclass
class LogAxis(ContinuousAxis):

    ticker: LogTicker = ...

    formatter: LogTickFormatter = ...

@dataclass
class CategoricalAxis(Axis, SeparatorLine, GroupText, SubgroupText):

    ticker: CategoricalTicker = ...

    formatter: CategoricalTickFormatter = ...

    group_label_orientation: LabelOrientation | float = ...

    subgroup_label_orientation: LabelOrientation | float = ...

@dataclass
class DatetimeAxis(LinearAxis):

    ticker: DatetimeTicker = ...

    formatter: DatetimeTickFormatter = ...

@dataclass
class MercatorAxis(LinearAxis):

    ticker: MercatorTicker = ...

    formatter: MercatorTickFormatter = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Any, Sequence

# Bokeh imports
from ..core.enums import (
    JitterRandomDistributionType as JitterRandomDistribution,
    StepModeType as StepMode,
)
from ..core.has_props import abstract
from ..model import Model
from .ranges import Range
from .sources import ColumnarDataSource

@abstract
@dataclass(init=False)
class Transform(Model):
    ...

@dataclass
class CustomJSTransform(Transform):

    args: dict[str, Any] = ...

    func: str = ...

    v_func: str = ...

@dataclass
class Dodge(Transform):

    value: float = ...

    range: Range | None = ...

@dataclass
class Jitter(Transform):

    mean: float = ...

    width: float = ...

    distribution: JitterRandomDistribution = ...

    range: Range | None = ...

@abstract
@dataclass(init=False)
class Interpolator(Transform):

    x: str | Sequence[float] | None = ...

    y: str | Sequence[float] | None = ...

    data: ColumnarDataSource | None = ...

    clip: bool = ...

@dataclass
class LinearInterpolator(Interpolator):
    ...

@dataclass
class StepInterpolator(Interpolator):

    mode: StepMode = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import ClassVar, Literal

# Bokeh imports
from .._types import CoordinateLike
from ..core.has_props import abstract
from ..model import Model
from .glyphs import Glyph
from .renderers import GlyphRenderer

ImplicitTarget = Literal["viewport", "canvas", "plot", "frame", "parent"]

@dataclass
class BoxNodes:

    def __init__(self, target: Model | ImplicitTarget) -> None: ...

    @property
    def left(self) -> Node: ...

    @property
    def right(self) -> Node: ...

    @property
    def top(self) -> Node: ...

    @property
    def bottom(self) -> Node: ...

    @property
    def top_left(self) -> Node: ...

    @property
    def top_center(self) -> Node: ...

    @property
    def top_right(self) -> Node: ...

    @property
    def center_left(self) -> Node: ...

    @property
    def center(self) -> Node: ...

    @property
    def center_right(self) -> Node: ...

    @property
    def bottom_left(self) -> Node: ...

    @property
    def bottom_center(self) -> Node: ...

    @property
    def bottom_right(self) -> Node: ...

    @property
    def width(self) -> Node: ...

    @property
    def height(self) -> Node: ...

@abstract
@dataclass(init=False)
class Coordinate(Model):
    ...

@dataclass
class XY(Coordinate):

    x: CoordinateLike = ...

    y: CoordinateLike = ...

@dataclass
class Indexed(Coordinate):

    index: int

    renderer: GlyphRenderer[Glyph]

@dataclass
class Node(Coordinate):

    target: Model | ImplicitTarget = ...

    symbol: str = ...

    offset: int = ...

    canvas: ClassVar[BoxNodes] = ...

    plot: ClassVar[BoxNodes] = ...

    frame: ClassVar[BoxNodes] = ...

    parent: ClassVar[BoxNodes] = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass

# Bokeh imports
from ..dom import DOMNode
from .ui_element import UIElement

@dataclass
class Pane(UIElement):

    elements: list[UIElement | DOMNode] = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

# Standard library imports
from dataclasses import dataclass
from typing import Sequence

# Bokeh imports
from ...core.enums import AutoType as Auto
from ...core.has_props import abstract
from ...model import Model
from ..css import Styles, StyleSheet
from ..nodes import Node
from .menus import Menu

@abstract
@dataclass(init=False)
class StyledElement(Model):

    html_attributes: dict[str, str] = ...

    html_id: str | None = ...

    css_classes: Sequence[str] = ...

    css_variables: dict[str, str | Node] = ...

    styles: dict[str, str | None] | Styles = ...

    stylesheets: list[StyleSheet | str | dict[str, dict[str, str | None] | Styles]] = ...

@abstract
@dataclass(init=False)
class UIElement(StyledElement):

    visible: bool = ...

    context_menu: Menu | Auto | None = ...



#-----------------------------------------------------------------------------
# Copyright (c) Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Common enumerations to be used together with |Enum| property.

This module provides many pre-defined enumerations, as well as functions
for creating new enumerations.

New enumerations can be created using the |enumeration| function:

.. code-block:: python

    #: Specify a nautically named side, port or starboard
    MyEnum = enumeration("port", "starboard")

Typically, enumerations are used to define |Enum| properties:

.. code-block:: python

    from bokeh.model import Model
    from bokeh.core.properties import Enum

    class MyModel(Model):

        location = Enum(MyEnum, help="""
        Whether the thing should be a port or starboard.
        """)

Enumerations have a defined order and support iteration:

.. code-block:: python

    >>> for loc in MyEnum:
    ...     print(loc)
    ...
    port
    starboard

as well as containment tests:

.. code-block:: python

    >>> "port" in MyEnum
    True

Enumerations can be easily documented in Sphinx documentation with the
:ref:`bokeh.sphinxext.bokeh_enum` Sphinx extension.

----

.. autofunction:: bokeh.core.enums.enumeration

----

.. |Enum| replace:: :class:`~bokeh.core.properties.Enum`
.. |enumeration| replace:: :func:`~bokeh.core.enums.enumeration`

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import (
    Any,
    Iterator,
    Literal,
    get_args,
)

# Bokeh imports
from .. import colors, palettes
from ..util.strings import nice_join

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'Align',
    'AlternationPolicy',
    'Anchor',
    'AngleUnits',
    'Auto',
    'AutosizeMode',
    'BuiltinFormatter',
    'ButtonType',
    'CalendarPosition',
    'ContextWhich',
    'CoordinateUnits',
    'DashPattern',
    'DateFormat',
    'DatetimeUnits',
    'Dimension',
    'Dimensions',
    'Direction',
    'FlowMode',
    'FontStyle',
    'HAlign',
    'HatchPattern',
    'HatchPatternAbbreviation',
    'HexTileOrientation',
    'HoldPolicy',
    'HorizontalLocation',
    'ImageOrigin',
    'ImplicitTarget',
    'JitterRandomDistribution',
    'KeyModifier',
    'LabelOrientation',
    'LatLon',
    'LegendClickPolicy',
    'LegendLocation',
    'LineCap',
    'LineDash',
    'LineJoin',
    'Location',
    'MapType',
    'MarkerType',
    'Movable',
    'NamedColor',
    'NumeralLanguage',
    'Orientation',
    'OutlineShapeName',
    'OutputBackend',
    'PaddingUnits',
    'Palette',
    'PanDirection',
    'Place',
    'RadiusDimension',
    'RegionSelectionMode',
    'RenderLevel',
    'ResetPolicy',
    'Resizable',
    'ResolutionType',
    'RoundingFunction',
    'ScrollbarPolicy',
    'SelectionMode',
    'SizingMode',
    'SizingPolicy',
    'SortDirection',
    'SpatialUnits',
    'StartEnd',
    'StepMode',
    'TeXDisplay',
    'TextAlign',
    'TextBaseline',
    'TextureRepetition',
    'ToolIcon',
    'ToolName',
    'TooltipAttachment',
    'TooltipFieldFormatter',
    'TrackPolicy',
    'VAlign',
    'VerticalAlign',
    'VerticalLocation',
    'WindowAxis',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

class Enumeration:
    ''' Represent an enumerated collection of values.

    .. note::
        Instances of ``Enumeration`` typically should not be constructed
        directly. Instead, use the |enumeration| function.

    '''
    __slots__ = ()

    _values: list[str]
    _default: str
    _case_sensitive: bool
    _quote: bool

    def __iter__(self) -> Iterator[str]:
        return iter(self._values)

    def __contains__(self, value: str) -> bool:
        if not self._case_sensitive:
            value = value.lower()
        return value in self._values

    def __str__(self) -> str:
        fn = repr if self._quote else str
        return f"Enumeration({', '.join(fn(x) for x in self._values)})"

    def __len__(self) -> int:
        return len(self._values)

    __repr__ = __str__

def enumeration(*values: Any, case_sensitive: bool = True, quote: bool = False) -> Enumeration:
    ''' Create an |Enumeration| object from a sequence of values.

    Call ``enumeration`` with a sequence of (unique) strings to create an
    Enumeration object:

    .. code-block:: python

        #: Specify the horizontal alignment for rendering text
        TextAlign = enumeration("left", "right", "center")

    Args:
        values (str) : string enumeration values, passed as positional arguments

            The order of arguments is the order of the enumeration, and the
            first element will be considered the default value when used
            to create |Enum| properties.

        case_sensitive (bool, optional) :
            Whether validation should consider case or not (default: True)

        quote (bool, optional) :
            Whether values should be quoted in the string representations
            (default: False)

    Raises:
        ValueError if values empty, if any value is not a string or not unique

    Returns:
        Enumeration

    '''
    if len(values) == 1 and hasattr(values[0], "__args__"):
        values = get_args(values[0])

    if not (values and all(isinstance(value, str) and value for value in values)):
        raise ValueError(f"expected a non-empty sequence of strings, got {nice_join(values)}")

    if len(values) != len(set(values)):
        raise ValueError(f"enumeration items must be unique, got {nice_join(values)}")

    attrs: dict[str, Any] = {value: value for value in values}
    attrs.update({
        "_values": list(values),
        "_default": values[0],
        "_case_sensitive": case_sensitive,
        "_quote": quote,
    })

    return type("Enumeration", (Enumeration,), attrs)()

#: Alignment (vertical or horizontal) of a child item
AlignType = Literal["start", "center", "end"]
Align = enumeration(AlignType)

#: Horizontal alignment of a child item
HAlignType = Literal["left", "center", "right"]
HAlign = enumeration(HAlignType)

#: Vertical alignment of a child item
VAlignType = Literal["top", "center", "bottom"]
VAlign = enumeration(VAlignType)

#: Specify to which items apply styling in a container (e.g. in a legend)
AlternationPolicyType = Literal["none", "even", "odd", "every"]
AlternationPolicy = enumeration(AlternationPolicyType)

#: Specify an anchor position on a box/frame
AnchorType = Literal[
    "top_left",    "top_center",    "top_right",
    "center_left", "center_center", "center_right",
    "bottom_left", "bottom_center", "bottom_right",
    "top", "left", "center", "right", "bottom",
]
Anchor = enumeration(AnchorType)

#: Specify the units for an angle value
AngleUnitsType = Literal["deg", "rad", "grad", "turn"]
AngleUnits = enumeration(AngleUnitsType)

#:
AutoType = Literal["auto"]
Auto = enumeration(AutoType)

#: Specify autosize mode for DataTable
AutosizeModeType = Literal["fit_columns", "fit_viewport", "force_fit", "none"]
AutosizeMode = enumeration(AutosizeModeType)

#: Names of built-in value formatters
BuiltinFormatterType = Literal["raw", "basic", "numeral", "printf", "datetime"]
BuiltinFormatter = enumeration(BuiltinFormatterType)

#: Specify a style for button widgets
ButtonTypeType = Literal["default", "primary", "success", "warning", "danger", "light"]
ButtonType = enumeration(ButtonTypeType)

#: Specify a position for the DatePicker calendar to display
CalendarPositionType = Literal["auto", "above", "below"]
CalendarPosition = enumeration(CalendarPositionType)

#: Specify which tick to add additional context to
ContextWhichType = Literal["start", "center", "end", "all"]
ContextWhich = enumeration(ContextWhichType)

#: Specify units for mapping coordinates
CoordinateUnitsType = Literal["canvas", "screen", "data"]
CoordinateUnits = enumeration(CoordinateUnitsType)

#: Specify a named dashing pattern for stroking lines
DashPatternType = Literal["solid", "dashed", "dotted", "dotdash", "dashdot"]
DashPattern = enumeration(DashPatternType)

#: Specify a format for printing dates
DateFormatType = Literal[
    "ATOM", "W3C", "RFC-3339", "ISO-8601", "COOKIE", "RFC-822",
    "RFC-850", "RFC-1036", "RFC-1123", "RFC-2822", "RSS", "TIMESTAMP",
]
DateFormat = enumeration(DateFormatType)

#: Specify a date/time scale
DatetimeUnitsType = Literal[
    "microseconds", "milliseconds", "seconds", "minsec",
    "minutes", "hourmin", "hours", "days", "months", "years",
]
DatetimeUnits = enumeration(DatetimeUnitsType)

#: Specify a vertical/horizontal dimension
DimensionType = Literal["width", "height"]
Dimension = enumeration(DimensionType)

#: Specify a vertical/horizontal dimensions
DimensionsType = Literal["width", "height", "both"]
Dimensions = enumeration(DimensionsType)

#: Specify a stroke direction for circles, wedges, etc.
DirectionType = Literal["clock", "anticlock"]
Direction = enumeration(DirectionType)

#: Specify the flow behavior in CSS layouts.
FlowModeType = Literal["block", "inline"]
FlowMode = enumeration(FlowModeType)

#: Specify the font style for rendering text
FontStyleType = Literal["normal", "italic", "bold", "bold italic"]
FontStyle = enumeration(FontStyleType)

#: Specify one of the built-in patterns for hatching fills
HatchPatternType = Literal[
    "blank",
    "dot",
    "ring",
    "horizontal_line",
    "vertical_line",
    "cross",
    "horizontal_dash",
    "vertical_dash",
    "spiral",
    "right_diagonal_line",
    "left_diagonal_line",
    "diagonal_cross",
    "right_diagonal_dash",
    "left_diagonal_dash",
    "horizontal_wave",
    "vertical_wave",
    "criss_cross",
]
HatchPattern = enumeration(HatchPatternType)

#: Specify one of the built-in patterns for hatching fills with a one-letter abbreviation
#:
#: The abbreviations are mapped as follows:
#:
#: .. code-block:: none
#:
#:     " "  :  blank
#:     "."  :  dot
#:     "o"  :  ring
#:     "-"  :  horizontal_line
#:     "|"  :  vertical_line
#:     "+"  :  cross
#:     '"'  :  horizontal_dash
#:     ":"  :  vertical_dash
#:     "@"  :  spiral
#:     "/"  :  right_diagonal_line
#:     "\\" :  left_diagonal_line
#:     "x"  :  diagonal_cross
#:     ","  :  right_diagonal_dash
#:     "`"  :  left_diagonal_dash
#:     "v"  :  horizontal_wave
#:     ">"  :  vertical_wave
#:     "*"  :  criss_cross
HatchPatternAbbreviationType = Literal[" ", ".", "o", "-", "|", "+", '"', ":", "@", "/", "\\", "x", ",", "`", "v", ">", "*"]
HatchPatternAbbreviation = enumeration(HatchPatternAbbreviationType, quote=True)

#: The orientation of the hex tiles
HexTileOrientationType = Literal["pointytop", "flattop"]
HexTileOrientation = enumeration(HexTileOrientationType)

#: Specify whether events should be combined or collected as-is when a Document hold is in effect
HoldPolicyType = Literal["combine", "collect"]
HoldPolicy = enumeration(HoldPolicyType)

#: Specify a horizontal location in plot layouts
HorizontalLocationType = Literal["left", "right"]
HorizontalLocation = enumeration(HorizontalLocationType)

#: Defines the coordinate space within an image
ImageOriginType = Literal["bottom_left", "top_left", "bottom_right", "top_right"]
ImageOrigin = enumeration(ImageOriginType)

#: Implicitly defined target of a node
ImplicitTargetType = Literal["viewport", "canvas", "plot", "frame", "parent"]
ImplicitTarget = enumeration(ImplicitTargetType)

#: Specify a distribution to use for the Jitter class
JitterRandomDistributionType = Literal["uniform", "normal"]
JitterRandomDistribution = enumeration(JitterRandomDistributionType)

#: Keyboard modifier key used to configure tools or report in UI events
KeyModifierType = Literal["shift", "ctrl", "alt"]
KeyModifier = enumeration(KeyModifierType)

#: Specify how labels are oriented with respect to an axis
LabelOrientationType = Literal["horizontal", "vertical", "parallel", "normal"]
LabelOrientation = enumeration(LabelOrientationType)

#: Specify whether a dimension or coordinate is latitude or longitude
LatLonType = Literal["lat", "lon"]
LatLon = enumeration(LatLonType)

#: Specify how a legend should respond to click events
LegendClickPolicyType = Literal["none", "hide", "mute"]
LegendClickPolicy = enumeration(LegendClickPolicyType)

#: Specify a fixed location for a legend
LegendLocationType = AnchorType
LegendLocation = Anchor

#: Specify how stroked lines should be terminated
LineCapType = Literal["butt", "round", "square"]
LineCap = enumeration(LineCapType)

#: Specify a named dash pattern for stroking lines
LineDashType = Literal["solid", "dashed", "dotted", "dotdash", "dashdot"]
LineDash = enumeration(LineDashType)

#: Specify how stroked lines should be joined together
LineJoinType = Literal["miter", "round", "bevel"]
LineJoin = enumeration(LineJoinType)

#: Specify a location in plot layouts
LocationType = Literal["above", "below", "left", "right"]
Location = enumeration(LocationType)

#: Specify a style for a Google map
MapTypeType = Literal["satellite", "roadmap", "terrain", "hybrid"]
MapType = enumeration(MapTypeType)

#: Specify one of the built-in marker types
MarkerTypeType = Literal[
    "asterisk", "circle", "circle_cross", "circle_dot", "circle_x",
    "circle_y", "cross", "dash", "diamond", "diamond_cross", "diamond_dot",
    "dot", "hex", "hex_dot", "inverted_triangle", "plus", "square",
    "square_cross", "square_dot", "square_pin", "square_x", "star", "star_dot",
    "triangle", "triangle_dot", "triangle_pin", "x", "y",
]
MarkerType = enumeration(MarkerTypeType)

#: Indicates in which dimensions an object (a renderer or an UI element) can be moved.
MovableType = Literal["none", "x", "y", "both"]
Movable = enumeration(MovableType)

#: Specify one of the CSS4 named colors (https://www.w3.org/TR/css-color-4/#named-colors)
NamedColor = enumeration(*colors.named.__all__, case_sensitive=False)

#: Specify a locale for printing numeric values
NumeralLanguageType = Literal["be-nl", "chs", "cs", "da-dk", "de-ch", "de", "en",
                              "en-gb", "es-ES", "es", "et", "fi", "fr-CA", "fr-ch",
                              "fr", "hu", "it", "ja", "nl-nl", "pl", "pt-br",
                              "pt-pt", "ru", "ru-UA", "sk", "th", "tr", "uk-UA"]
NumeralLanguage = enumeration(NumeralLanguageType)

#: Specify a vertical/horizontal orientation for something
OrientationType = Literal["horizontal", "vertical"]
Orientation = enumeration(OrientationType)

#: Names of pre-defined outline shapes (used in ``Text.outline_shape``)
OutlineShapeNameType = Literal["none", "box", "rectangle", "square", "circle", "ellipse", "trapezoid", "parallelogram", "diamond", "triangle"]
OutlineShapeName = enumeration(OutlineShapeNameType)

#: Specify an output backend to render a plot area onto
OutputBackendType = Literal["canvas", "svg", "webgl"]
OutputBackend = enumeration(OutputBackendType)

#: Whether range padding should be interpreted a percentage or and absolute quantity
PaddingUnitsType = Literal["percent", "absolute"]
PaddingUnits = enumeration(PaddingUnitsType)

#: Which direction click pan tool acts on.
PanDirectionType = Literal["left", "right", "up", "down", "west", "east", "north", "south"]
PanDirection = enumeration(PanDirectionType)

#: Specify the name of a palette from :ref:`bokeh.palettes`
PaletteType = str # TODO
Palette = enumeration(*palettes.__palettes__)

#: Placement of a layout element, in particular in border-style layouts
PlaceType = Literal["above", "below", "left", "right", "center"]
Place = enumeration(PlaceType)

#: Specify which dimension or dimensions to use when measuring circle radius
RadiusDimensionType = Literal["x", "y", "max", "min"]
RadiusDimension = enumeration(RadiusDimensionType)

#: Specify a position in the render order for a renderer
RenderLevelType = Literal["image", "underlay", "glyph", "guide", "annotation", "overlay"]
RenderLevel = enumeration(RenderLevelType)

#: What reset actions should occur on a Plot reset
ResetPolicyType = Literal["standard", "event_only"]
ResetPolicy = enumeration(ResetPolicyType)

#: Indicates in which dimensions an object (a renderer or an UI element) can be resized.
ResizableType = Literal["none", "left", "right", "top", "bottom", "x", "y", "all"]
Resizable = enumeration(ResizableType)

#: Specify which resolutions should be used for stripping of leading zeros
ResolutionTypeType = Literal["microseconds", "milliseconds", "seconds", "minsec", "minutes", "hourmin", "hours", "days", "months", "years"]
ResolutionType = enumeration(ResolutionTypeType)

#: Specify a policy for  how numbers should be rounded
RoundingFunctionType = Literal["round", "nearest", "floor", "rounddown", "ceil", "roundup"]
RoundingFunction = enumeration(RoundingFunctionType)

#: Scrollbar policies
ScrollbarPolicyType = Literal["auto", "visible", "hidden"]
ScrollbarPolicy = enumeration(ScrollbarPolicyType)

#: Region selection modes
RegionSelectionModeType = Literal["replace", "append", "intersect", "subtract", "xor"]
RegionSelectionMode = enumeration(RegionSelectionModeType)

#: Selection modes
SelectionModeType = Literal[RegionSelectionModeType, "toggle"]
SelectionMode = enumeration(SelectionModeType)

#: Sizing mode policies
SizingModeType = Literal["stretch_width", "stretch_height", "stretch_both", "scale_width", "scale_height", "scale_both", "fixed", "inherit"]
SizingMode = enumeration(SizingModeType)

#: Individual sizing mode policies
SizingPolicyType = Literal["fixed", "fit", "min", "max"]
SizingPolicy = enumeration(SizingPolicyType)

#: Specify sorting directions
SortDirectionType = Literal["ascending", "descending"]
SortDirection = enumeration(SortDirectionType)

#: Specify units for mapping values
SpatialUnitsType = Literal["screen", "data"]
SpatialUnits = enumeration(SpatialUnitsType)

#: Specify a start/end value
StartEndType = Literal["start", "end"]
StartEnd = enumeration(StartEndType)

#: Specify a mode for stepwise interpolation
StepModeType = Literal["before", "after", "center"]
StepMode = enumeration(StepModeType)

#: Display mode in TeX
TeXDisplayType = Literal["inline", "block", "auto"]
TeXDisplay = enumeration(TeXDisplayType)

#: Specify the horizontal alignment for rendering text
TextAlignType = Literal["left", "right", "center"]
TextAlign = enumeration(TextAlignType)

#: Specify the baseline location for rendering text
TextBaselineType = Literal["top", "middle", "bottom", "alphabetic", "hanging", "ideographic"]
TextBaseline = enumeration(TextBaselineType)

#: Specify how textures used as canvas patterns should repeat
TextureRepetitionType = Literal["repeat", "repeat_x", "repeat_y", "no_repeat"]
TextureRepetition = enumeration(TextureRepetitionType)

#: Well known tool icon names
ToolIconType = Literal[
    "append_mode",
    "arrow_down_to_bar",
    "arrow_up_from_bar",
    "auto_box_zoom",
    "bold",
    "box_edit",
    "box_select",
    "box_zoom",
    "caret_down",
    "caret_left",
    "caret_right",
    "caret_up",
    "check",
    "chevron_down",
    "chevron_left",
    "chevron_right",
    "chevron_up",
    "clear_selection",
    "copy",
    "crosshair",
    "dark_theme",
    "delete",
    "freehand_draw",
    "fullscreen",
    "help",
    "hover",
    "intersect_mode",
    "invert_selection",
    "italic",
    "lasso_select",
    "light_theme",
    "line_edit",
    "maximize",
    "minimize",
    "pan",
    "pin",
    "point_draw",
    "pointer",
    "poly_draw",
    "poly_edit",
    "polygon_select",
    "range",
    "redo",
    "replace_mode",
    "reset",
    "save",
    "see_off",
    "see_on",
    "settings",
    "square",
    "square_check",
    "subtract_mode",
    "tap_select",
    "text_align_center",
    "text_align_left",
    "text_align_right",
    "undo",
    "unknown",
    "unpin",
    "wheel_pan",
    "wheel_zoom",
    "x_box_select",
    "x_box_zoom",
    "x_grip",
    "x_pan",
    "xor_mode",
    "y_box_select",
    "y_box_zoom",
    "y_grip",
    "y_pan",
    "zoom_in",
    "zoom_out",
]
ToolIcon = enumeration(ToolIconType)

#: Known tool names/aliases
ToolNameType = Literal[
  "auto_box_zoom",
  "box_select",
  "box_zoom",
  "click",
  "copy",
  "crosshair",
  "doubletap",
  "examine",
  "freehand_draw",
  "fullscreen",
  "help",
  "hover",
  "lasso_select",
  "pan",
  "pan_down",
  "pan_east",
  "pan_left",
  "pan_north",
  "pan_right",
  "pan_south",
  "pan_up",
  "pan_west",
  "poly_select",
  "redo",
  "reset",
  "save",
  "tap",
  "undo",
  "wheel_zoom",
  "xbox_select",
  "xbox_zoom",
  "xcrosshair",
  "xpan",
  "xwheel_pan",
  "xwheel_zoom",
  "xzoom_in",
  "xzoom_out",
  "ybox_select",
  "ybox_zoom",
  "ycrosshair",
  "ypan",
  "ywheel_pan",
  "ywheel_zoom",
  "yzoom_in",
  "yzoom_out",
  "zoom_in",
  "zoom_out",
]
ToolName = enumeration(ToolNameType)

#: Specify an attachment for tooltips
TooltipAttachmentType = Literal["horizontal", "vertical", "left", "right", "above", "below"]
TooltipAttachment = enumeration(TooltipAttachmentType)

#: Specify how a format string for a tooltip field should be interpreted
TooltipFieldFormatterType = Literal["numeral", "datetime", "printf"]
TooltipFieldFormatter = enumeration(TooltipFieldFormatterType)

#: Grid track (row/column) sizing policies
TrackPolicyType = Literal["auto", "min", "max", "flex", "fixed"]
TrackPolicy = enumeration(TrackPolicyType)

#: Specify the vertical alignment for rendering text
VerticalAlignType = Literal["top", "middle", "bottom"]
VerticalAlign = enumeration(VerticalAlignType)

#: Specify a vertical location in plot layouts
VerticalLocationType = Literal["above", "below"]
VerticalLocation = enumeration(VerticalLocationType)

#: Specify a which axis to use for windowed auto-ranging
WindowAxisType = Literal["none", "x", "y"]
WindowAxis = enumeration(WindowAxisType)

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------
